from typing import Set
from kalc.misc.util import cpuConvertToAbstractProblem, memConvertToAbstractProblem
from kalc.model.system.primitives import Label
from kalc.misc.object_factory import labelFactory
from kalc.misc.const import *

from poodle import Object

class ModularKind(Object):
    external_defaults = {}
    policy_engine = None
    @classmethod
    def register_property(cls, name, type, default):
        cls.__annotations__[name] = type
        cls.external_defaults[name] = default
 
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for prop_name, default_value in self.external_defaults.items():
            if default_value is None: continue
            setattr(self, prop_name, default_value)

    def __getattribute__(self, name):
        if name == "policy":
            if self.policy_engine:
                return self.policy_engine.get(self)
        return super().__getattribute__(name)


class HasLabel(Object):
    metadata_labels: Set[Label]
    metadata_name: str

class HasLimitsRequests(Object):
    """A mixin class to implement Limts/Requests loading and initialiaztion"""
    memRequest: int
    cpuRequest: int
    memLimit: int
    memLimitsStatus: StatusLim
    """Status to set if the limit is reached"""
    cpuLimit: int
    cpuLimitsStatus: StatusLim
    """Status to set if the limit is reached"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cpuLimit = -1
        self.memLimit = -1
        self.cpuRequest = -1
        self.memRequest = -1
        self.memLimitsStatus = STATUS_LIM["Limit Met"]
        self.cpuLimitsStatus = STATUS_LIM["Limit Met"]

    @property
    def spec_template_spec_containers__resources_limits_cpu(self):
        pass
    @spec_template_spec_containers__resources_limits_cpu.setter
    def spec_template_spec_containers__resources_limits_cpu(self, res):
        if self.cpuLimit == -1: self.cpuLimit = 0
        self.cpuLimit += cpuConvertToAbstractProblem(res)

    @property
    def spec_template_spec_containers__resources_limits_memory(self):
        pass
    @spec_template_spec_containers__resources_limits_memory.setter
    def spec_template_spec_containers__resources_limits_memory(self, res):
        if self.memLimit == -1: self.memLimit = 0
        self.memLimit += memConvertToAbstractProblem(res)

    @property
    def spec_template_spec_containers__resources_requests_cpu(self):
        pass
    @spec_template_spec_containers__resources_requests_cpu.setter
    def spec_template_spec_containers__resources_requests_cpu(self, res):
        if self.cpuRequest == -1: self.cpuRequest = 0
        self.cpuRequest += cpuConvertToAbstractProblem(res)

    @property
    def spec_template_spec_containers__resources_requests_memory(self):
        pass
    @spec_template_spec_containers__resources_requests_memory.setter
    def spec_template_spec_containers__resources_requests_memory(self, res):
        if self.memRequest == -1: self.memRequest = 0
        self.memRequest += memConvertToAbstractProblem(res)



from typing import Set
from kalc.model.kinds.Pod import Pod
from kalc.model.kinds.Node import Node
from kalc.model.kinds.Service import Service
from kalc.model.system.Scheduler import Scheduler
from kalc.model.system.globals import GlobalVar
from kalc.misc.const import *
from poodle import planned
from kalc.policy import policy_engine, BasePolicy

from kalc.model.system.Scheduler import Scheduler
from kalc.model.system.globals import GlobalVar
from kalc.model.kinds.Service import Service
from kalc.model.kinds.Node import Node
from kalc.model.kinds.Pod import Pod
from kalc.model.kinds.Deployment import Deployment
from kalc.model.kinds.DaemonSet import DaemonSet
from kalc.model.kinds.PriorityClass import PriorityClass
from kalc.model.kubernetes import KubernetesCluster
from kalc.misc.const import *


class BalancedClusterPolicy(BasePolicy):
    TYPE = "property"

    def register(self):
        GlobalVar.register_property(name="deploymentsWithAntiaffinityBalanced", type=bool, default=False)
        GlobalVar.register_property(name="add_node_enabled", type=bool, default=False)

    def set(self, val: bool):
        assert isinstance(val, bool), "Can only accept True or False"

        if val:
            # enable

            def hypothesis_do_rebalance_2_pods():
                # TODO: hypotheses can not work in parallel this way: will modify main object
                self.register_goal(self.target_object.deploymentsWithAntiaffinityBalanced, "==", True)

            def hypothesis_do_rebalance_allow_add_node():
                # TODO: hypotheses can not work in parallel this way: will modify main object
                self.target_object.add_node_enabled = True
                self.register_goal(self.target_object.deploymentsWithAntiaffinityBalanced, "==", True)
            
            self.register_hypothesis("2 pods rebalanced", hypothesis_do_rebalance_2_pods, order=1)
            self.register_hypothesis("Added node to help with pods rebalancing", hypothesis_do_rebalance_allow_add_node, order=10)
        else:
            # disable
            pass

    @planned(cost=1)
    def mark_checked_pod_as_antiaffinity_checked_for_target_pod(self,
        target_pod: Pod,
        antiaffinity_pod: Pod,
        globalVar: GlobalVar,
        scheduler: Scheduler,
        antiaffinity_pod_node: Node):
        if antiaffinity_pod.atNode != target_pod.atNode and \
                    target_pod.antiaffinity_set == True and \
                antiaffinity_pod not in target_pod.calc_antiaffinity_pods_list:
                target_pod.calc_antiaffinity_pods_list.add(antiaffinity_pod)
                target_pod.calc_antiaffinity_pods_list_length += 1
        assert antiaffinity_pod in target_pod.podsMatchedByAntiaffinity 
        assert antiaffinity_pod.atNode == antiaffinity_pod_node
        assert antiaffinity_pod_node.isNull == False
        # assert globalVar.block_policy_calculated == True
        globalVar.block_policy_calculated = True


    @planned(cost=1)
    def calculate_length_of_nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod(self,
        target_pod: Pod,
        antiaffinity_pod: Pod,
        antiaffinity_pod_node: Node,
        debug_nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod_length: int):
        assert antiaffinity_pod in target_pod.podsMatchedByAntiaffinity
        assert antiaffinity_pod in target_pod.calc_affinity_pods_list
        assert antiaffinity_pod_node == antiaffinity_pod.atNode
        if antiaffinity_pod_node not in target_pod.nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod:
            assert debug_nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod_length == target_pod.nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod_length           
            target_pod.nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod.add(antiaffinity_pod_node)
            target_pod.nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod_length += 1


    @planned(cost=1)
    def mark_checked_pod_as_affinity_checked_for_target_pod(self,
        target_pod: Pod,
        checked_pod: Pod,
        globalVar: GlobalVar,
        scheduler: Scheduler):
        if checked_pod.atNode == target_pod.atNode and \
                    target_pod.affinity_set == True and \
                checked_pod not in target_pod.calc_affinity_pods_list:
                target_pod.calc_affinity_pods_list.add(checked_pod)
                target_pod.calc_affinity_pods_list_length += 1
        assert checked_pod in target_pod.podsMatchedByAffinity 
        # assert globalVar.block_policy_calculated == True 
        globalVar.block_policy_calculated = True


    @planned(cost=1)
    def mark_that_node_cant_allocate_pod_by_cpu(self,
        pod: Pod,
        node: Node,
        globalVar: GlobalVar):
        if not node in pod.nodesThatCantAllocateThisPod:
            assert pod.cpuRequest > node.cpuCapacity - node.currentFormalCpuConsumption
            pod.nodesThatCantAllocateThisPod.add(node)
            pod.nodesThatCantAllocateThisPod_length += 1
        # assert globalVar.block_policy_calculated == True
        globalVar.block_policy_calculated = True
        assert node.isNull == False


    @planned(cost=1)
    def mark_that_node_cant_allocate_pod_by_mem(self,
        pod: Pod,
        node: Node,
        globalVar: GlobalVar):
        assert node.isNull == False
        if not node in pod.nodesThatCantAllocateThisPod:
            assert pod.memRequest > node.memCapacity - node.currentFormalMemConsumption
            pod.nodesThatCantAllocateThisPod.add(node)
            pod.nodesThatCantAllocateThisPod_length += 1
        # assert globalVar.block_policy_calculated == True
        globalVar.block_policy_calculated = True

  
    @planned(cost=1)
    def mark_antiaffinity_met_because_all_antiaffinity_pods_are_matched(self,
        pod: Pod,
        globalVar: GlobalVar):
        assert pod.calc_antiaffinity_pods_list_length == pod.podsMatchedByAntiaffinity_length
        assert pod.antiaffinity_set == True
        # assert globalVar.block_policy_calculated == True
        pod.antiaffinity_met = True
        globalVar.block_policy_calculated = True


    @planned(cost=1)
    def mark_affinity_met_because_all_affinity_pods_are_matched(self,
        pod: Pod,
        globalVar: GlobalVar):
        assert pod.calc_affinity_pods_list_length == pod.podsMatchedByAffinity_length
        assert pod.affinity_set == True
        # assert globalVar.block_policy_calculated == True
        pod.affinity_met = True
        globalVar.block_policy_calculated = True


    @planned(cost=1)
    def mark_antiaffinity_met_because_all_antiaffinity_pods_are_matched_and_those_that_cant_dont_suite(self,
        pod: Pod,
        globalVar: GlobalVar):
        assert pod.calc_antiaffinity_pods_list_length == pod.target_number_of_antiaffinity_pods
        assert pod.antiaffinity_set == True
        # assert globalVar.block_policy_calculated == True
        pod.antiaffinity_met = True
        globalVar.block_policy_calculated = True


    # @planned(cost=1)
    def mark_antiaffinity_met_because_all_antiaffinity_pods_are_matched_and_those_that_cant_dont_suite_below_the_limit_for_node_amount(self,
        pod: Pod,
        globalVar: GlobalVar):
        assert pod.nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod_length + pod.nodesThatCantAllocateThisPod_length == globalVar.amountOfNodes_limit
        assert pod.antiaffinity_set == True
        pod.antiaffinity_met = True

    
    @planned(cost=1)
    def Set_antiaffinity_between_pods_of_deployment(self,
        pod1: Pod,
        pod2: Pod,
        deployment: Deployment,
        globalVar: GlobalVar):
        if pod1 not in pod2.podsMatchedByAffinity and pod2 not in pod1.podsMatchedByAffinity:
            assert pod1 in deployment.podList
            assert pod2 in deployment.podList
            assert deployment.NumberOfPodsOnSameNodeForDeployment == globalVar.maxNumberOfPodsOnSameNodeForDeployment
            pod1.antiaffinity_set = True
            pod1.podsMatchedByAffinity.add(pod2)
            pod1.podsMatchedByAffinity_length += 1
            pod2.antiaffinity_set = True
            pod2.podsMatchedByAffinity.add(pod1)
            pod2.podsMatchedByAffinity_length += 1

 
    @planned(cost=1)
    def Reduce_maxNumberOfPodsOnSameNodeForDeployment(self,
        globalVar: GlobalVar):
        globalVar.maxNumberOfPodsOnSameNodeForDeployment -= 1       
   

    @planned(cost=1)
    def Calculate_fulfilled_antiaffinity_pods_of_deployment(self,
        pod1: Pod,
        pod2: Pod,
        deployment: Deployment):
        assert pod1 in deployment.podList
        assert deployment.amountOfActivePods > 0
        assert pod1.podsMatchedByAffinity_length == deployment.amountOfActivePods
        deployment.amountOfPodsWithAntiaffinity += 1
    

    @planned(cost=1)
    def Calculate_fulfilled_antiaffinity_pods_of_deployment_with_limited_number_of_pods(self,
        pod1: Pod,
        pod2: Pod,
        deployment: Deployment,
        globalVar: GlobalVar):
        assert pod1 in deployment.podList
        assert globalVar.target_amountOfPodsWithAntiaffinity > 0
        assert pod1.podsMatchedByAffinity_length == globalVar.target_amountOfPodsWithAntiaffinity
        deployment.amountOfPodsWithAntiaffinity += 1


    @planned(cost=1)
    def mark_finalization_of_antiaffinity_setting_for_pods_of_deployment(self,
        deployment: Deployment,
        globalVar: GlobalVar):
        assert deployment.amountOfPodsWithAntiaffinity == deployment.amountOfActivePods
        globalVar.amountOfDeploymentsWithAntiaffinity += 1
    

    @planned(cost=1)
    def mark_finalization_of_antiaffinity_setting_for_pods_of_deployment(self,
        deployment: Deployment,
        globalVar: GlobalVar):
        assert globalVar.amountOfDeploymentsWithAntiaffinity == globalVar.target_amountOfDeploymentsWithAntiaffinity
        globalVar.deploymentsWithAntiaffinityBalanced = True


    @planned(cost=1)
    def Add_node(self,
            node : Node,
            globalVar: GlobalVar):
        assert globalVar.add_node_enabled == True
        assert node.status == STATUS_NODE["New"]
        node.status = STATUS_NODE["Active"]
        globalVar.amountOfNodes += 1
    
policy_engine.register(Service, "self_antiaffinity", BalancedClusterPolicy)




from kalc.model.system.base import ModularKind
from kalc.model.system.Controller import Controller
from kalc.model.system.base import HasLimitsRequests
from kalc.model.kinds.PriorityClass import PriorityClass, zeroPriorityClass
from kalc.model.system.Scheduler import Scheduler
import kalc.model.kinds.Pod as mpod
from kalc.model.kinds.ReplicaSet import ReplicaSet
from kalc.model.system.primitives import Status, Label
from kalc.misc.const import STATUS_POD, STATUS_SCHED, StatusDeployment
import kalc.model.kinds.Node as mnode
from poodle import *
from typing import Set
from logzero import logger
import kalc.misc.util as util
import random
import yaml, copy, jsonpatch, difflib
from kalc.misc.metrics import calculate_maxNumberOfPodsOnSameNode_metrics

class YAMLable():
    yaml_orig: {}
    yaml: {}
    patchJSON: []
    rawYaml: str

    def set_yaml_nested_key(self, yamlmod, keys, value = None):
        l = len(keys)
        yaml = yamlmod
        if l > 1 or value == None:
            for key in range(l-1):
                if not keys[key] in yaml :
                    yaml[keys[key]] = {}
                yaml = yaml[keys[key]]
        if value != None and not keys[l-1] in yaml:
            yaml[keys[l-1]] = value
    


    def get_patch(self):

        if hasattr(self, "yaml") and hasattr(self, "yaml_orig"):
            orig = yaml.dump(self.yaml_orig).splitlines(keepends=True)
            new = yaml.dump(self.yaml).splitlines(keepends=True)
            return "".join(list(difflib.unified_diff(orig, new, n=4)))  #use differ n to make more reliable diff file
        return ""

class Deployment(ModularKind, Controller, HasLimitsRequests, YAMLable):
    spec_replicas: int
    metadata_name: str
    metadata_namespace: str
    apiVersion: str
    amountOfActivePods: int
    status: StatusDeployment
    podList: Set["mpod.Pod"]
    spec_template_spec_priorityClassName: str
    searchable: bool
    hash: str
    NumberOfPodsOnSameNodeForDeployment: int
    calc_PodsWithAntiaffinitySet: Set["mpod.Pod"]
    calc_PodsWithAntiaffinitySet_length: int
    podsMatchedByAntiaffinity_length: int
    



    def __init__(self, *args, **kwargs):
        super().__init__( *args, **kwargs)
        #TODO fill pod-template-hash with https://github.com/kubernetes/kubernetes/blob/0541d0bb79537431421774465721f33fd3b053bc/pkg/controller/controller_utils.go#L1024
        self.metadata_name = "modelDeployment"+str(random.randint(100000000, 999999999))
        # self.metadata_name = "model-default-name"
        self.hash = ''.join(random.choice("0123456789abcdef") for i in range(8))
        self.amountOfActivePods = 0
        self.searchable = False
        self.spec_template_spec_priorityClassName = "Normal-zero"
        self.priorityClass = zeroPriorityClass
        self.spec_replicas = 0
        self.NumberOfPodsOnSameNodeForDeployment = 1
        self.amountOfPodsWithAntiaffinity = 5
        self.calc_PodsWithAntiaffinitySet_length = 0
        self.podsMatchedByAntiaffinity_length = 0

    def hook_after_create(self, object_space):
        deployments = filter(lambda x: isinstance(x, Deployment), object_space)
        for deploymentController in deployments:
            if str(deploymentController.metadata_name) == str(self.metadata_name):
                message = "Error from server (AlreadyExists): deployments.{0} \"{1}\" already exists".format(str(self.apiVersion).split("/")[0], self.metadata_name)
                logger.error(message)
                raise AssertionError(message)
        self.create_pods(object_space, self.spec_replicas._get_value())
        calculate_maxNumberOfPodsOnSameNode_metrics(self,object_space)


    

    def create_pods(self, object_space, replicas, start_from=0):
        scheduler = next(filter(lambda x: isinstance(x, Scheduler), object_space))
        for replicaNum in range(replicas):
            new_pod = mpod.Pod()
            hash1 = self.hash
            hash2 = str(replicaNum+start_from)
            new_pod.metadata_name = "{0}-{1}-{2}".format(str(self.metadata_name),hash1,hash2)
            for label in self.spec_selector_matchLabels._get_value():
                if not (label in new_pod.metadata_labels._get_value()):
                    new_pod.metadata_labels.add(label)
            new_pod.metadata_labels.add(Label("pod-template-hash:{0}".format(hash1)))
            new_pod.cpuRequest = self.cpuRequest
            new_pod.memRequest = self.memRequest
            new_pod.cpuLimit = self.cpuLimit
            new_pod.memLimit = self.memLimit
            new_pod.status = STATUS_POD["Pending"]
            new_pod.hook_after_load(object_space, _ignore_orphan=True) # for service<>pod link
            new_pod.set_priority(object_space, self)
            new_pod.hasDeployment = True
            self.podList.add(new_pod)
            # self.check_pod(new_pod, object_space)
            object_space.append(new_pod)

    def hook_after_load(self, object_space):
        deployments = filter(lambda x: isinstance(x, Deployment), object_space)
        for deploymentController in deployments:
            if deploymentController != self and str(deploymentController.metadata_name) == str(self.metadata_name):
                message = "Error from server (AlreadyExists): deployments.{0} \"{1}\" already exists".format(str(self.apiVersion).split("/")[0], self.metadata_name)
                logger.error(message)
                raise AssertionError(message)
        pods = filter(lambda x: isinstance(x, mpod.Pod), object_space)
        replicasets = filter(lambda x: isinstance(x, ReplicaSet), object_space)
        #look for ReplicaSet with corresonding owner reference
        for replicaset in replicasets:
            br=False
            if util.getint(replicaset.spec_replicas) == 0: continue # ignore as it is likely history object
            if str(replicaset.metadata_ownerReferences__name) == str(self.metadata_name):
                for pod_template_hash in list(replicaset.metadata_labels._get_value()):
                    if str(pod_template_hash).split(":")[0] == "pod-template-hash":
                        self.hash = str(pod_template_hash).split(":")[1]
                        br = True
                        break
            if br: break

        for pod in pods:
            br = False
            # look for right pod-template-hash
            for pod_template_hash in list(pod.metadata_labels._get_value()):
                if str(pod_template_hash).split(":")[0] == "pod-template-hash" and str(pod_template_hash).split(":")[1] == self.hash :
                    self.podList.add(pod)
                    pod.hasDeployment = True
                    br = True
            if br and pod.status._get_value() == "Running":
                self.amountOfActivePods += 1
                    # self.check_pod(pod, object_space)
        calculate_maxNumberOfPodsOnSameNode_metrics(self,object_space)

    def hook_scale_before_create(self, object_space, new_replicas):
        self.spec_replicas = new_replicas

    def hook_after_apply(self, object_space):
        deployments = filter(lambda x: isinstance(x, Deployment), object_space)
        old_deployment = self
        for deploymentController in deployments:
            if deploymentController != self and str(deploymentController.metadata_name) == str(self.metadata_name):
                old_deployment = deploymentController
                break
        # if old DEployment not found
        if old_deployment == self:
            self.hook_after_create(object_space)
        else:
            self.podList = old_deployment.podList # copy pods
            self.hook_scale_after_load(object_space, old_deployment.spec_replicas._get_value()) # extend or trimm pods
            object_space.remove(old_deployment) # delete old Deployment

    #Call me only atfter loading this Controller
    def hook_scale_after_load(self, object_space, old_replicas):
        diff_replicas = self.spec_replicas._get_value() - old_replicas
        if diff_replicas == 0:
            logger.warning("Nothing to scale. You try to scale deployment {0} for the same replicas value {1}".format(self.metadata_name, self.spec_replicas))
        if diff_replicas < 0:
            #remove pods
            for _ in range(diff_replicas * -1):
                pod = self.podList._get_value().pop(-1)
                object_space.remove(pod)
                util.objRemoveByName(self.podList._get_value(), pod.metadata_name)
        if diff_replicas > 0:
            self.create_pods(object_space, diff_replicas, self.spec_replicas._get_value())
        #scale memory and cpu
        for pod in util.objDeduplicatorByName(self.podList._get_value()):
            pod.cpuRequest = self.cpuRequest
            pod.memRequest = self.memRequest
            pod.cpuLimit = self.cpuLimit
            pod.memLimit = self.memLimit
            pod.set_priority(object_space, self)

    def affinity_required_handler(self, label = None, node = None, antiAffinity = True):
        assert node is None, "todo nodes not supported"
        if not hasattr(self, "yaml"):
            self.yaml = {}
        json_orig = copy.deepcopy(self.yaml)
        if not hasattr(self, "patchJSON"):
            self.patchJSON = []
        if antiAffinity:
            podAntiAffinityType = 'podAntiAffinity'
        else:
            podAntiAffinityType = 'podAffinity'

        selector = 'podSelector'
        selectorValue = "{0}-{1}".format(podAntiAffinityType,''.join(random.choice("0123456789abcdef") for i in range(8)))
        
        if not label is None:
            selector = label['key']
            selectorValue = label['value']
        #append affinity
        self.set_yaml_nested_key(yamlmod = self.yaml, keys=['spec', 'template', 'spec', podAntiAffinityType, 'requiredDuringSchedulingIgnoredDuringExecution'], value=[])
        labelSelector = {}
        labelSelector["matchExpressions"]=[]
        matchExpr = {}
        matchExpr['key']= selector
        matchExpr['operator']= 'In'
        matchExpr['values'] = []
        matchExpr['values'].append(selectorValue)
        labelSelector["matchExpressions"].append(matchExpr)
        self.yaml['spec']['template']['spec'][podAntiAffinityType]['requiredDuringSchedulingIgnoredDuringExecution'].append({"labelSelector": labelSelector})
        if label is None:
            #appent pod template selector
            self.set_yaml_nested_key(yamlmod = self.yaml, keys=['spec', 'template', 'metadata','labels', selector], value = selectorValue)
            #append top selector
            self.set_yaml_nested_key(yamlmod = self.yaml, keys=['spec','selector','matchLabels', selector], value = selectorValue)
        self.patchJSON.extend(jsonpatch.make_patch(json_orig, self.yaml))

    def scale_replicas_handler(self, replicas):
        if not hasattr(self, "yaml"):
            self.yaml = {}
        if not hasattr(self, "yaml_orig"):
            self.yaml = {}
        json_orig = copy.deepcopy(self.yaml)
        if not hasattr(self, "patchJSON"):
            self.patchJSON = []
        self.set_yaml_nested_key(yamlmod = self.yaml, keys=['spec','replicas'], value=replicas)
        self.yaml['spec']['replicas']=replicas
        self.patchJSON.extend(jsonpatch.make_patch(json_orig, self.yaml))

        
    def __str__(self): return str(self.metadata_name)


import random
from kalc.model.system.Controller import Controller
from kalc.model.system.base import HasLimitsRequests, ModularKind
from kalc.model.kinds.PriorityClass import PriorityClass
from kalc.model.system.Scheduler import Scheduler
import kalc.model.kinds.Pod as mpod
from kalc.model.system.primitives import Status
from kalc.misc.const import STATUS_POD, STATUS_SCHED
from poodle import *
from typing import Set
from logzero import logger

class ReplicaSet(ModularKind, Controller, HasLimitsRequests):
    metadata_name: str
    metadata_namespace: str
    apiVersion: str
    metadata_ownerReferences__kind: str
    metadata_ownerReferences__name: str
    spec_replicas: int
   
    def __init__(self, *args, **kwargs):
        super().__init__( *args, **kwargs)
        #TODO fill pod-template-hash with https://github.com/kubernetes/kubernetes/blob/0541d0bb79537431421774465721f33fd3b053bc/pkg/controller/controller_utils.go#L1024
        self.metadata_name = "modelReplicaSet"+str(random.randint(100000000, 999999999))
        # self.metadata_name = "model-default-name"
        # self.hash = "superhash"
        self.hash = str(random.randint(100000000, 999999999))
        self.spec_replicas = 0
        self.metadata_ownerReferences__kind = "NONE"
        self.metadata_ownerReferences__name = "NONE"


    def hook_after_create(self, object_space):
        pass

    def hook_after_load(self, object_space):
        pass


from poodle import Object, planned
from typing import Set
from kalc.misc.const import *
import kalc.model.kinds.Pod as mpod
from kalc.model.system.primitives import StatusSched
import sys


class Scheduler(Object):
    queueLength: int
    status: StatusSched
    podQueue: Set["mpod.Pod"]
    podQueue_excluded_pods_length: int
    podQueue_excluded_pods: Set["mpod.Pod"]
    # debug_var: bool
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.queueLength = 0
        # self.debug_var = False
        self.status = STATUS_SCHED["Clean"]
        self.podQueue_excluded_pods_length = 0

    def __str__(self): return str(self._get_value())


import sys
from poodle import planned
from logzero import logger
from kalc.model.system.base import HasLimitsRequests, HasLabel
from kalc.model.system.Scheduler import Scheduler
from kalc.model.system.globals import GlobalVar
from kalc.model.system.primitives import TypeServ
from kalc.model.system.math import Permutations, permutation_list
from kalc.model.system.Controller import Controller
from kalc.model.system.primitives import Label
from kalc.model.kinds.Service import Service
from kalc.model.kinds.Deployment import Deployment
from kalc.model.kinds.DaemonSet import DaemonSet
from kalc.model.kinds.Pod import Pod
from kalc.model.kinds.Node import Node
from kalc.model.kinds.PriorityClass import PriorityClass, zeroPriorityClass
from kalc.misc.const import *
from kalc.model.kubeactions import KubernetesModel
from kalc.misc.util import cpuConvertToAbstractProblem, memConvertToAbstractProblem
from kalc.misc.const import STATUS_SCHED
from kalc.misc.script_generator import script_remove_node
import re
import itertools

class ExcludeDict:
    name: str
    objType: str
    obj: TypeServ

    def __init__(self, kn):
        self.objType = kn.split(":")[0]
        self.name = kn.split(":")[1]
        self.obj = TypeServ(self.name)

class K8ServiceInterruptSearch(KubernetesModel):

    @planned(cost=1)
    def NodeNServiceInterupted(self,globalVar:GlobalVar, scheduler: Scheduler):
        assert globalVar.is_node_disrupted == True
        assert globalVar.is_service_disrupted == True
        assert scheduler.status == STATUS_SCHED["Clean"]
        globalVar.goal_achieved = True 
        
    @planned(cost=1)
    def Mark_node_outage_event(self,
        node:"Node",
        globalvar:GlobalVar):
        assert node.status == STATUS_NODE["Inactive"]
        assert node.searchable == True
        globalvar.is_node_disrupted = True
        
        
def mark_excluded(object_space, excludeStr, skip_check=False):
    exclude = []
    if excludeStr != None:
        for kn in excludeStr.split(","):
            exclude.append(ExcludeDict(kn))
    else: 
        return
    names = []
    types = []
    for obj in object_space:
        if hasattr(obj, 'metadata_name'):
            names.append(str(obj.metadata_name))
            types.append(str(obj.__class__.__name__))
            for objExclude in exclude:
                re_name = "^" + objExclude.name.replace('*', '.*') + "$"
                re_objType =  "^" + objExclude.objType.replace('*', '.*') + "$"
                if (re.search(re.compile(re_objType), str(obj.__class__.__name__)) is not None) and \
                    (re.search(re.compile(re_name), str(obj.metadata_name)) is not None):
                    # print("mark unserchable ", str(obj.__class__.__name__), ":", str(obj.metadata_name ))
                    obj.searchable = False
    if skip_check : return
    for objExclude in exclude:
        re_name = "^" + objExclude.name.replace('*', '.*') + "$"
        re_objType =  "^" + objExclude.objType.replace('*', '.*') + "$"

        typeCheck = True
        for type_name in types:  
            if re.search(re.compile(re_objType), type_name) is not None:
                typeCheck = False
        if typeCheck:
            raise AssertionError("Error: no such type '{0}'".format(objExclude.objType))

        nameCheck = True
        for metadata_name in names:
            if re.search(re.compile(re_name), metadata_name) is not None:
                nameCheck = False
        if nameCheck:
            raise AssertionError("Error: no such {1}: '{0}'".format(objExclude.name, objExclude.objType))

class OptimisticRun(K8ServiceInterruptSearch):
    goal = lambda self: self.globalVar.goal_achieved == True 
    
    # @planned(cost=9000) # this works for deployment-outage case
    # def SchedulerQueueCleanHighCost(self, scheduler: Scheduler, global_: GlobalVar):
    #     assert scheduler.status == STATUS_SCHED["Clean"]
    #     assert global_.block_node_outage_in_progress == False
    #     global_.goal_achieved = True
    
  
    @planned(cost=1)
    def Scheduler_cant_place_pod(self, scheduler: "Scheduler",
        globalVar: GlobalVar):
        # assert globalVar.block_node_outage_in_progress == False
        scheduler.queueLength -= 1
        
class Check_deployments(OptimisticRun):
    @planned(cost=1)
    def AnyDeploymentInterrupted(self,globalVar:GlobalVar,
                scheduler: "Scheduler"):
        assert globalVar.is_deployment_disrupted == True
        # assert scheduler.status == STATUS_SCHED["Clean"]
        globalVar.goal_achieved = True 
        
    @planned(cost=1)
    def MarkDeploymentOutageEvent(self,
                deployment_current: Deployment,
                pod_current: Pod,
                global_: "GlobalVar",
                scheduler: "Scheduler"
            ):
        assert scheduler.status == STATUS_SCHED["Clean"] 
        assert deployment_current.amountOfActivePods < deployment_current.spec_replicas
        assert deployment_current.searchable == True
        assert pod_current in  deployment_current.podList
        assert pod_current.status == STATUS_POD["Pending"]

        deployment_current.status = STATUS_DEPLOYMENT["Interrupted"]
        global_.is_deployment_disrupted = True
        
class Check_services(OptimisticRun):
    @planned(cost=1)
    def MarkServiceOutageEvent(self,
                service1: Service,
                pod1: Pod,
                global_: "GlobalVar",
                scheduler: "Scheduler"
            ):
            
        assert scheduler.status == STATUS_SCHED["Clean"] 
        assert service1.amountOfActivePods == 0
        # assert service1.status == STATUS_SERV["Started"] # TODO: Activate  this condition -  if service has to be started before eviction  
        assert service1.searchable == True  
        assert pod1 in service1.podList
        assert service1.isNull == False

        service1.status = STATUS_SERV["Interrupted"]
        global_.is_service_disrupted = True #TODO:  Optimistic search 
        
class Check_services_restart(OptimisticRun):
    @planned(cost=1)
    def MarkServiceOutageEvent(self,
                service1: Service,
                pod1: Pod,
                global_: "GlobalVar",
                scheduler: "Scheduler"
            ):
            
        # assert scheduler.status == STATUS_SCHED["Clean"] # Removed this assert  to make profile that check if service outage may happed temporary ( restart of service on the other node will help)
        assert service1.amountOfActivePods == 0
        # assert service1.status == STATUS_SERV["Started"] # TODO: Activate  this condition -  if service has to be started before eviction  
        assert service1.searchable == True  
        assert pod1 in service1.podList
        assert service1.isNull == False

        service1.status = STATUS_SERV["Interrupted"]
        global_.is_service_disrupted = True #TODO:  Optimistic search 
        
    @planned(cost=1) # this works for no-outage case
    def SchedulerQueueCleanLowCost(self, scheduler: Scheduler, global_: GlobalVar):
        assert scheduler.status == STATUS_SCHED["Clean"]
        assert global_.block_node_outage_in_progress == False
        global_.goal_achieved = True

    @planned(cost=1)
    def AnyServiceInterrupted(self,globalVar:GlobalVar, scheduler: Scheduler):
        assert globalVar.is_service_disrupted == True
        assert scheduler.status == STATUS_SCHED["Clean"]
        globalVar.goal_achieved = True 
class Check_daemonsets(OptimisticRun):        
    @planned(cost=1)
    def MarkDaemonsetOutageEvent(self,
                daemonset_current: DaemonSet,
                pod_current: Pod,
                global_: "GlobalVar",
                scheduler: "Scheduler"
            ):
        assert scheduler.status == STATUS_SCHED["Clean"] 
        assert daemonset_current.searchable == True
        assert pod_current in  daemonset_current.podList
        assert pod_current.status == STATUS_POD["Pending"]

        # daemonset_current.status = STATUS_DAEMONSET_INTERRUPTED
        global_.is_daemonset_disrupted = True
        
class CheckNodeOutage(Check_services):
    goal = lambda self: self.globalVar.goal_achieved == True and \
                                self.globalVar.is_node_disrupted == True
class Check_services_and_deployments(Check_services,Check_deployments):
    pass

class Check_services_deployments_daemonsets(Check_daemonsets,Check_services,Check_deployments):
    pass
class Check_node_outage_and_service_restart(Check_services_restart):
    goal = lambda self: self.globalVar.is_service_disrupted == True and \
                                self.globalVar.is_node_disrupted == True


class HypothesisysClean(K8ServiceInterruptSearch):
    def Remove_pod_common_part(self,
        pod: Pod,
        scheduler: Scheduler):
        pod.status = STATUS_POD["Outaged"]
        scheduler.podQueue.remove(pod)
        scheduler.queueLength -= 1
        scheduler.queueLength -= 0 #TODO: remove this once replaced with costs
        scheduler.queueLength -= 0 #TODO: remove this once replaced with costs

    @planned(cost=100)
    def Remove_pod_from_the_cluster(self,
                service : Service,
                pod : Pod,
                scheduler : Scheduler,
                globalVar : GlobalVar
            ):
        # This action helps to remove pods from queue 
        assert globalVar.block_node_outage_in_progress == False
        assert pod.status == STATUS_POD["Pending"]
        assert pod in scheduler.podQueue
        if pod.hasService == True:
            assert pod in service.podList
            if service.amountOfActivePods + service.amountOfPodsInQueue == 1:
                    service.status = STATUS_SERV["Interrupted"]
                    globalVar.is_service_disrupted = True
                    self.Remove_pod_common_part()
            else:
                assert service.amountOfActivePods + service.amountOfPodsInQueue > 1
                self.Remove_pod_common_part()
        else:
            assert pod.hasService == False
            self.Remove_pod_common_part()
    
class HypothesisysNodeAndService(HypothesisysClean):
    goal = lambda self: self.scheduler.status == STATUS_SCHED["Clean"] and \
                        self.globalVar.is_node_disrupted == True and \
                        self.globalVar.is_service_disrupted == True


class HypothesisysNode(HypothesisysClean):
    goal = lambda self: self.scheduler.status == STATUS_SCHED["Clean"] and \
                           self.globalVar.is_node_disrupted == True 

# class Antiaffinity_implement(KubernetesModel):

#     def generate_goal(self):
#         self.generated_goal_in = []
#         self.generaged_goal_eq = []
#         for service in filter(lambda s: isinstance(s, Service) and s.antiaffinity == True , self.objectList):
#             for pod1, pod2 in itertools.permutations(filter(lambda x: isinstance(x, Pod) and x in service.podList, self.objectList),2):
#                 self.generated_goal_in.append([pod1, pod2.not_on_same_node])

#     def goal(self):
#         for what, where in self.generated_goal_in:
#             assert what in where
#         for what1, what2 in self.generaged_goal_eq:
#             assert what1 == what2
       

#     @planned(cost=1)
#     def manually_initiate_killing_of_pod(self,
#         node_with_outage: "Node",
#         pod_killed: "podkind.Pod",
#         globalVar: GlobalVar
#         ):
#         assert pod_killed.status == STATUS_POD["Running"]
#         pod_killed.status = STATUS_POD["Killing"]
#     @planned(cost=1)
#     def Not_at_same_node(self,
#             pod1: Pod,
#             pod2: Pod,
#             node_of_pod2: Node,
#             scheduler: Scheduler):
#         assert node_of_pod2 == pod2.atNode
#         assert pod1.atNode in node_of_pod2.different_than
#         assert scheduler.status == STATUS_SCHED["Clean"]
#         pod1.not_on_same_node.add(pod2)


# class Antiaffinity_implement_with_add_node(Antiaffinity_implement):

#     @planned(cost=1)
#     def Add_node(self,
#                 node : Node):
#         assert node.status == STATUS_NODE["New"]
#         node.status = STATUS_NODE["Active"]

# class Antiaffinity_prefered(KubernetesModel):

#     goal = lambda self: self.globalVar.antiaffinity_prefered_policy_met == True

#     @planned(cost=1)
#     def mark_antiaffinity_prefered_policy_set(self,
#         service: Service,
#         globalVar: GlobalVar):
#         assert service.amountOfPodsInAntiaffinityGroup == service.targetAmountOfPodsOnDifferentNodes
#         assert service.isSearched == True
#         assert service.antiaffinity == True
#         service.antiaffinity_prefered_policy_set = True

#     @planned(cost=1)
#     def mark_antiaffinity_prefered_policy_met(self,
#         service: Service,
#         globalVar: GlobalVar):
#         assert service.amountOfPodsOnDifferentNodes == service.targetAmountOfPodsOnDifferentNodes
#         assert service.isSearched == True
#         assert service.antiaffinity == True
#         service.antiaffinity_prefered_policy_met = True

#     @planned(cost=1)
#     def manually_initiate_killing_of_pod_3(self,
#         node_with_outage: "Node",
#         pod_killed: "podkind.Pod",
#         globalVar: GlobalVar
#         ):
#         assert pod_killed.status == STATUS_POD["Running"]
#         pod_killed.status = STATUS_POD["Killing"]

#     @planned(cost=1)
#     def mark_2_pods_of_service_as_not_at_same_node(self,
#             pod1: Pod,
#             pod2: Pod,
#             node_of_pod1: Node,
#             node_of_pod2: Node,
#             service: Service,
#             scheduler: Scheduler):
#         assert node_of_pod2 == pod2.atNode
#         assert pod1.atNode in node_of_pod2.different_than
#         assert pod1 in service.podList
#         assert pod2 in service.podList
#         assert scheduler.status == STATUS_SCHED["Clean"]
#         pod1.not_on_same_node.add(pod2)
#         service.amountOfPodsOnDifferentNodes = 2

#     @planned(cost=1)
#     def mark_3_pods_of_service_as_not_at_same_node(self,
#             pod1: Pod,
#             pod2: Pod,
#             pod3: Pod,
#             node_of_pod1: Node,
#             node_of_pod2: Node,
#             node_of_pod3: Node,
#             service: Service,
#             scheduler: Scheduler):
        
#         assert node_of_pod1 == pod1.atNode
#         assert node_of_pod2 == pod2.atNode
#         assert node_of_pod3 == pod3.atNode
#         assert node_of_pod1 in node_of_pod2.different_than
#         assert node_of_pod1 in node_of_pod3.different_than
#         assert node_of_pod2 in node_of_pod3.different_than
#         assert pod1 in service.podList
#         assert pod2 in service.podList
#         assert pod3 in service.podList
#         assert scheduler.status == STATUS_SCHED["Clean"]
#         service.amountOfPodsOnDifferentNodes = 3

#     @planned(cost=1)
#     def mark_4_pods_of_service_as_not_at_same_node(self,
#             pod1: Pod,
#             pod2: Pod,
#             pod3: Pod,
#             pod4: Pod,
#             node_of_pod1: Node,
#             node_of_pod2: Node,
#             node_of_pod3: Node,
#             node_of_pod4: Node,
#             service: Service,
#             scheduler: Scheduler):
        
#         assert node_of_pod1 == pod1.atNode
#         assert node_of_pod2 == pod2.atNode
#         assert node_of_pod3 == pod3.atNode
#         assert node_of_pod4 == pod4.atNode
#         assert node_of_pod1 in node_of_pod2.different_than
#         assert node_of_pod1 in node_of_pod3.different_than
#         assert node_of_pod1 in node_of_pod4.different_than
#         assert node_of_pod2 in node_of_pod3.different_than
#         assert node_of_pod2 in node_of_pod4.different_than
#         assert node_of_pod3 in node_of_pod4.different_than
#         assert pod1 in service.podList
#         assert pod2 in service.podList
#         assert pod3 in service.podList
#         assert pod4 in service.podList
#         assert scheduler.status == STATUS_SCHED["Clean"]
#         service.amountOfPodsOnDifferentNodes = 4


#     @planned(cost=1)
#     def mark_5_pods_of_service_as_not_at_same_node(self,
#             pod1: Pod,
#             pod2: Pod,
#             pod3: Pod,
#             pod4: Pod,
#             pod5: Pod,
#             node_of_pod1: Node,
#             node_of_pod2: Node,
#             node_of_pod3: Node,
#             node_of_pod4: Node,
#             node_of_pod5: Node,
#             service: Service,
#             scheduler: Scheduler):
        
#         assert node_of_pod1 == pod1.atNode
#         assert node_of_pod2 == pod2.atNode
#         assert node_of_pod3 == pod3.atNode
#         assert node_of_pod4 == pod4.atNode
#         assert node_of_pod5 == pod5.atNode
#         assert node_of_pod1 in node_of_pod2.different_than
#         assert node_of_pod1 in node_of_pod3.different_than
#         assert node_of_pod1 in node_of_pod4.different_than
#         assert node_of_pod1 in node_of_pod5.different_than
#         assert node_of_pod2 in node_of_pod3.different_than
#         assert node_of_pod2 in node_of_pod4.different_than
#         assert node_of_pod2 in node_of_pod5.different_than
#         assert node_of_pod3 in node_of_pod4.different_than
#         assert node_of_pod3 in node_of_pod5.different_than
#         assert node_of_pod4 in node_of_pod5.different_than

#         assert pod1 in service.podList
#         assert pod2 in service.podList
#         assert pod3 in service.podList
#         assert pod4 in service.podList
#         assert pod5 in service.podList
#         assert scheduler.status == STATUS_SCHED["Clean"]
#         service.amountOfPodsOnDifferentNodes = 5

# class Antiaffinity_prefered_with_add_node(Antiaffinity_prefered):
#     @planned(cost=1)
#     def Add_node(self,
#             node : Node):
#         assert node.status == STATUS_NODE["New"]
#         node.status = STATUS_NODE["Active"]

class Antiaffinity_set(KubernetesModel):

    goal = lambda self: self.service.antiaffinity_prefered_policy_set == True

    @planned(cost=1)
    def mark_antiaffinity_set(self,
        pod1: Pod,
        pod2: Pod):
        assert pod2.isSearched == True
        assert pod1 in pod2.podsMatchedByAntiaffinity
        pod1.antiaffinity_set = True

class Antiaffinity_met(KubernetesModel):
    @planned(cost=1)
    def mark_antiaffinity_met(self,
        pod: Pod):
        assert pod.calc_antiaffinity_pods_list_length == pod.target_number_of_antiaffinity_pods
        pod.antiaffinity_met = True
    @planned(cost=1)
    def mark_antiaffinity_preferred_met(self,
        pod: Pod):
        assert pod.calc_antiaffinity_preferred_pods_list_length == pod.target_number_of_antiaffinity_preferred_pods
        pod.antiaffinity_preferred_met = True
    
    def generate_goal(self):
        self.generated_goal_in = []
        self.generated_goal_eq = []
        for target_pod in filter(lambda p: isinstance(p, Pod) and p.antiaffinity_set == True, self.objectList):
                self.generated_goal_eq.append([target_pod.calc_antiaffinity_pods_list_length, target_pod.target_number_of_antiaffinity_pods])
        for target_pod in filter(lambda p: isinstance(p, Pod) and p.antiaffinity_preferred_set == True, self.objectList):
                self.generated_goal_eq.append([target_pod.calc_antiaffinity_pods_list_length, target_pod.target_number_of_antiaffinity_preferred_pods])

    def goal(self):
        for what, where in self.generated_goal_in:
            assert what in where
        for what1, what2 in self.generated_goal_eq:
            assert what1 == what2

    @planned(cost=1)
    def manually_initiate_killing_of_pod(self,
        node_with_outage: "Node",
        pod_killed: "podkind.Pod",
        globalVar: GlobalVar
        ):
        assert pod_killed.status == STATUS_POD["Running"]
        pod_killed.status = STATUS_POD["Killing"]

    @planned(cost=1)
    def Not_at_same_node_preferred(self,
            pod1: Pod,
            pod2: Pod,
            node_of_pod2: Node,
            scheduler: Scheduler):
        assert pod2 in pod1.podsMatchedByAntiaffinityPrefered
        assert pod2 not in pod1.calc_antiaffinity_preferred_pods_list
        assert node_of_pod2 == pod2.atNode
        assert pod1.atNode in node_of_pod2.different_than
        assert scheduler.status == STATUS_SCHED["Clean"]
        pod1.not_on_same_node.add(pod2)
        pod1.calc_antiaffinity_preferred_pods_list.add(pod2)
        pod1.calc_antiaffinity_preferred_pods_list_length += 1
        
class Antiaffinity_check_basis(KubernetesModel):
    # @planned(cost=1)
    # def brakepoint(self,
    #     target_pod: Pod):
    #     assert target_pod.status == STATUS_POD["Pending"]
    #     # assert target_pod.antiaffinity_set == True
    #     target_pod.antiaffinity_met = True
    # @planned(cost=1)
    # def brakepoint2(self,
    #     target_pod: Pod,
    #     node: Node):
    #     assert target_pod.status == STATUS_POD["Running"]
    #     assert target_pod.atNode == node
    #     assert node.isSearched == True
    #     target_pod.antiaffinity_met = True
    # @planned(cost=1)
    # def finish_cluster_changes_calculate_meassures(self,
    #     scheduler: Scheduler,
    #     globalVar: GlobalVar):
    #     assert globalVar.block_policy_calculated == False
    #     assert scheduler.queueLength == 0
    #     globalVar.block_policy_calculated = True
    @planned(cost=1)
    def mark_checked_pod_as_antiaffinity_checked_for_target_pod(self,
        target_pod: Pod,
        antiaffinity_pod: Pod,
        globalVar: GlobalVar,
        scheduler: Scheduler,
        antiaffinity_pod_node: Node):
        if antiaffinity_pod.atNode != target_pod.atNode and \
                    target_pod.antiaffinity_set == True and \
                antiaffinity_pod not in target_pod.calc_antiaffinity_pods_list:
                target_pod.calc_antiaffinity_pods_list.add(antiaffinity_pod)
                target_pod.calc_antiaffinity_pods_list_length += 1
        assert antiaffinity_pod in target_pod.podsMatchedByAntiaffinity
        assert globalVar.deploymentsWithAntiaffinityBalanced == True 
        assert antiaffinity_pod.atNode == antiaffinity_pod_node
        assert antiaffinity_pod_node.isNull == False
        # assert globalVar.block_policy_calculated == True
        globalVar.block_policy_calculated = True
    # @planned(cost=1)
    # def calculate_length_of_nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod(self,
    #     target_pod: Pod,
    #     antiaffinity_pod: Pod,
    #     antiaffinity_pod_node: Node,
    #     debug_nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod_length: int):
    #     assert antiaffinity_pod in target_pod.podsMatchedByAntiaffinity
    #     assert antiaffinity_pod in target_pod.calc_affinity_pods_list
    #     assert antiaffinity_pod_node == antiaffinity_pod.atNode
    #     if antiaffinity_pod_node not in target_pod.nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod:
    #         assert debug_nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod_length == target_pod.nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod_length           
    #         target_pod.nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod.add(antiaffinity_pod_node)
    #         target_pod.nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod_length += 1
    # @planned(cost=1)
    # def mark_checked_pod_as_affinity_checked_for_target_pod(self,
    #     target_pod: Pod,
    #     checked_pod: Pod,
    #     globalVar: GlobalVar,
    #     scheduler: Scheduler):
    #     if checked_pod.atNode == target_pod.atNode and \
    #                 target_pod.affinity_set == True and \
    #             checked_pod not in target_pod.calc_affinity_pods_list:
    #             target_pod.calc_affinity_pods_list.add(checked_pod)
    #             target_pod.calc_affinity_pods_list_length += 1
    #     assert checked_pod in target_pod.podsMatchedByAffinity
    #     assert globalVar.deploymentsWithAntiaffinityBalanced == True 
    #     # assert globalVar.block_policy_calculated == True 
    #     globalVar.block_policy_calculated = True
    @planned(cost=1)
    def mark_2_pods_of_deployment_as_not_at_same_node(self,
            pod1: Pod,
            pod2: Pod,
            node_of_pod1: Node,
            node_of_pod2: Node,
            deployment: Deployment):
        # assert node_of_pod2 == pod2.atNode
        assert pod1.atNode in node_of_pod2.different_than
        assert pod1 in deployment.podList
        assert pod2 in deployment.podList
        assert pod1.antiaffinity_set == True 
        assert pod2.antiaffinity_set == True
        assert pod2 in pod1.podsMatchedByAntiaffinity
        assert pod1 in pod2.podsMatchedByAntiaffinity
        if pod2 not in pod1.calc_antiaffinity_pods_list and \
            pod1 not in pod2.calc_antiaffinity_pods_list:
            pod1.calc_antiaffinity_pods_list.add(pod2)
            pod1.calc_antiaffinity_pods_list_length += 1
            pod2.calc_antiaffinity_pods_list.add(pod1)
            pod2.calc_antiaffinity_pods_list_length += 1

    @planned(cost=1)
    def mark_3_pods_of_deployment_as_not_at_same_node(self,
            pod1: Pod,
            pod2: Pod,
            pod3: Pod,
            node_of_pod1: Node,
            node_of_pod2: Node,
            node_of_pod3: Node,
            deployment: Deployment):
        
        # assert node_of_pod1 == pod1.atNode
        # assert node_of_pod2 == pod2.atNode
        # assert node_of_pod3 == pod3.atNode
        assert node_of_pod1 in node_of_pod2.different_than
        assert node_of_pod1 in node_of_pod3.different_than
        assert node_of_pod2 in node_of_pod3.different_than
        assert pod1 in deployment.podList
        assert pod2 in deployment.podList
        assert pod3 in deployment.podList
        assert pod1.antiaffinity_set == True 
        assert pod2.antiaffinity_set == True
        assert pod3.antiaffinity_set == True
        assert pod2 in pod1.podsMatchedByAntiaffinity
        assert pod1 in pod2.podsMatchedByAntiaffinity
        assert pod2 in pod3.podsMatchedByAntiaffinity
        assert pod1 in pod3.podsMatchedByAntiaffinity
        assert pod3 in pod1.podsMatchedByAntiaffinity
        assert pod3 in pod2.podsMatchedByAntiaffinity

        if pod2 not in pod1.calc_antiaffinity_pods_list and \
            pod1 not in pod2.calc_antiaffinity_pods_list and \
            pod2 not in pod3.calc_antiaffinity_pods_list and \
            pod1 not in pod3.calc_antiaffinity_pods_list and \
            pod3 not in pod1.calc_antiaffinity_pods_list and \
            pod3 not in pod2.calc_antiaffinity_pods_list:

            pod1.calc_antiaffinity_pods_list.add(pod2)
            pod1.calc_antiaffinity_pods_list.add(pod3)
            pod1.calc_antiaffinity_pods_list_length += 2
            pod2.calc_antiaffinity_pods_list.add(pod1)
            pod2.calc_antiaffinity_pods_list.add(pod3)
            pod2.calc_antiaffinity_pods_list_length += 2
            pod3.calc_antiaffinity_pods_list.add(pod1)
            pod3.calc_antiaffinity_pods_list.add(pod2)
            pod3.calc_antiaffinity_pods_list_length += 2


# THis can't be used yet because of  BUG 131
    # @planned(cost=1)
    # def mark_4_pods_of_deployment_as_not_at_same_node(self,
    #         pod1: Pod,
    #         pod2: Pod,
    #         pod3: Pod,
    #         pod4: Pod,
    #         node_of_pod1: Node,
    #         node_of_pod2: Node,
    #         node_of_pod3: Node,
    #         node_of_pod4: Node,
    #         deployment: Deployment):
        
    #     assert node_of_pod1 == pod1.atNode
    #     assert node_of_pod2 == pod2.atNode
    #     assert node_of_pod3 == pod3.atNode
    #     assert node_of_pod4 == pod4.atNode
    #     assert node_of_pod1 in node_of_pod2.different_than
    #     assert node_of_pod1 in node_of_pod3.different_than
    #     assert node_of_pod1 in node_of_pod4.different_than
    #     assert node_of_pod2 in node_of_pod3.different_than
    #     assert node_of_pod2 in node_of_pod4.different_than
    #     assert node_of_pod3 in node_of_pod4.different_than
    #     assert pod1 in deployment.podList
    #     assert pod2 in deployment.podList
    #     assert pod3 in deployment.podList
    #     assert pod4 in deployment.podList
    #     assert pod1.antiaffinity_set == True 
    #     assert pod2.antiaffinity_set == True
    #     assert pod3.antiaffinity_set == True
    #     assert pod4.antiaffinity_set == True
    #     assert pod1 in pod2.podsMatchedByAntiaffinity
    #     assert pod1 in pod3.podsMatchedByAntiaffinity
    #     assert pod1 in pod4.podsMatchedByAntiaffinity
    #     assert pod2 in pod1.podsMatchedByAntiaffinity
    #     assert pod2 in pod3.podsMatchedByAntiaffinity
    #     assert pod2 in pod4.podsMatchedByAntiaffinity
    #     assert pod3 in pod1.podsMatchedByAntiaffinity
    #     assert pod3 in pod2.podsMatchedByAntiaffinity
    #     assert pod3 in pod4.podsMatchedByAntiaffinity
    #     assert pod4 in pod1.podsMatchedByAntiaffinity
    #     assert pod4 in pod2.podsMatchedByAntiaffinity
    #     assert pod4 in pod3.podsMatchedByAntiaffinity
        
    #     if pod1 not in pod2.calc_antiaffinity_pods_list and \
    #         pod1 not in pod3.calc_antiaffinity_pods_list and \
    #         pod1 not in pod4.calc_antiaffinity_pods_list and \
    #         pod2 not in pod1.calc_antiaffinity_pods_list and \
    #         pod2 not in pod3.calc_antiaffinity_pods_list and \
    #         pod2 not in pod4.calc_antiaffinity_pods_list and \
    #         pod3 not in pod1.calc_antiaffinity_pods_list and \
    #         pod3 not in pod2.calc_antiaffinity_pods_list and \
    #         pod3 not in pod4.calc_antiaffinity_pods_list and \
    #         pod4 not in pod1.calc_antiaffinity_pods_list and \
    #         pod4 not in pod2.calc_antiaffinity_pods_list and \
    #         pod4 not in pod3.calc_antiaffinity_pods_list:

    #         pod1.calc_antiaffinity_pods_list.add(pod2)
    #         pod1.calc_antiaffinity_pods_list.add(pod3)
    #         pod1.calc_antiaffinity_pods_list.add(pod4)
    #         pod1.calc_antiaffinity_pods_list_length += 3
    #         pod2.calc_antiaffinity_pods_list.add(pod1)
    #         pod2.calc_antiaffinity_pods_list.add(pod3)
    #         pod2.calc_antiaffinity_pods_list.add(pod4)
    #         pod2.calc_antiaffinity_pods_list_length += 3
    #         pod3.calc_antiaffinity_pods_list.add(pod1)
    #         pod3.calc_antiaffinity_pods_list.add(pod2)
    #         pod3.calc_antiaffinity_pods_list.add(pod4)
    #         pod3.calc_antiaffinity_pods_list_length += 3
    #         pod4.calc_antiaffinity_pods_list.add(pod1)
    #         pod4.calc_antiaffinity_pods_list.add(pod2)
    #         pod4.calc_antiaffinity_pods_list.add(pod3)
    #         pod4.calc_antiaffinity_pods_list_length += 3


# THis can't be used yet because of  BUG 131
    # @planned(cost=1)
    # def mark_5_pods_of_deployment_as_not_at_same_node(self,
    #         pod1: Pod,
    #         pod2: Pod,
    #         pod3: Pod,
    #         pod4: Pod,
    #         pod5: Pod,
    #         node_of_pod1: Node,
    #         node_of_pod2: Node,
    #         node_of_pod3: Node,
    #         node_of_pod4: Node,
    #         node_of_pod5: Node,
    #         deployment: Deployment):
        
    #     assert node_of_pod1 == pod1.atNode
    #     assert node_of_pod2 == pod2.atNode
    #     assert node_of_pod3 == pod3.atNode
    #     assert node_of_pod4 == pod4.atNode
    #     assert node_of_pod5 == pod5.atNode
    #     assert node_of_pod1 in node_of_pod2.different_than
    #     assert node_of_pod1 in node_of_pod3.different_than
    #     assert node_of_pod1 in node_of_pod4.different_than
    #     assert node_of_pod1 in node_of_pod5.different_than
    #     assert node_of_pod2 in node_of_pod3.different_than
    #     assert node_of_pod2 in node_of_pod4.different_than
    #     assert node_of_pod2 in node_of_pod5.different_than
    #     assert node_of_pod3 in node_of_pod4.different_than
    #     assert node_of_pod3 in node_of_pod5.different_than
    #     assert node_of_pod4 in node_of_pod5.different_than

    #     assert pod1 in deployment.podList
    #     assert pod2 in deployment.podList
    #     assert pod3 in deployment.podList
    #     assert pod4 in deployment.podList
    #     assert pod5 in deployment.podList
    #     assert pod1.antiaffinity_set == True 
    #     assert pod2.antiaffinity_set == True
    #     assert pod3.antiaffinity_set == True
    #     assert pod4.antiaffinity_set == True
        
    #     assert pod1 in pod2.podsMatchedByAntiaffinity
    #     assert pod1 in pod3.podsMatchedByAntiaffinity
    #     assert pod1 in pod4.podsMatchedByAntiaffinity
    #     assert pod1 in pod5.podsMatchedByAntiaffinity
    #     assert pod2 in pod1.podsMatchedByAntiaffinity
    #     assert pod2 in pod3.podsMatchedByAntiaffinity
    #     assert pod2 in pod4.podsMatchedByAntiaffinity
    #     assert pod2 in pod5.podsMatchedByAntiaffinity
    #     assert pod3 in pod1.podsMatchedByAntiaffinity
    #     assert pod3 in pod2.podsMatchedByAntiaffinity
    #     assert pod3 in pod4.podsMatchedByAntiaffinity
    #     assert pod3 in pod5.podsMatchedByAntiaffinity
    #     assert pod4 in pod1.podsMatchedByAntiaffinity
    #     assert pod4 in pod2.podsMatchedByAntiaffinity
    #     assert pod4 in pod3.podsMatchedByAntiaffinity
    #     assert pod4 in pod5.podsMatchedByAntiaffinity
    #     assert pod5 in pod1.podsMatchedByAntiaffinity
    #     assert pod5 in pod2.podsMatchedByAntiaffinity
    #     assert pod5 in pod3.podsMatchedByAntiaffinity
    #     assert pod5 in pod4.podsMatchedByAntiaffinity

    #     if pod1 not in pod2.calc_antiaffinity_pods_list  and \
    #         pod1 not in pod3.calc_antiaffinity_pods_list and \
    #         pod1 not in pod4.calc_antiaffinity_pods_list and \
    #         pod1 not in pod5.calc_antiaffinity_pods_list and \
    #         pod2 not in pod1.calc_antiaffinity_pods_list and \
    #         pod2 not in pod3.calc_antiaffinity_pods_list and \
    #         pod2 not in pod4.calc_antiaffinity_pods_list and \
    #         pod2 not in pod5.calc_antiaffinity_pods_list and \
    #         pod3 not in pod1.calc_antiaffinity_pods_list and \
    #         pod3 not in pod2.calc_antiaffinity_pods_list and \
    #         pod3 not in pod4.calc_antiaffinity_pods_list and \
    #         pod3 not in pod5.calc_antiaffinity_pods_list and \
    #         pod4 not in pod1.calc_antiaffinity_pods_list and \
    #         pod4 not in pod2.calc_antiaffinity_pods_list and \
    #         pod4 not in pod3.calc_antiaffinity_pods_list and \
    #         pod4 not in pod5.calc_antiaffinity_pods_list and \
    #         pod5 not in pod1.calc_antiaffinity_pods_list and \
    #         pod5 not in pod2.calc_antiaffinity_pods_list and \
    #         pod5 not in pod3.calc_antiaffinity_pods_list and \
    #         pod5 not in pod4.calc_antiaffinity_pods_list:

    #         pod1.calc_antiaffinity_pods_list.add(pod2)
    #         pod1.calc_antiaffinity_pods_list.add(pod3)
    #         pod1.calc_antiaffinity_pods_list.add(pod4)
    #         pod1.calc_antiaffinity_pods_list.add(pod5)
    #         pod1.calc_antiaffinity_pods_list_length += 4

    #         pod2.calc_antiaffinity_pods_list.add(pod1)
    #         pod2.calc_antiaffinity_pods_list.add(pod3)
    #         pod2.calc_antiaffinity_pods_list.add(pod4)
    #         pod2.calc_antiaffinity_pods_list.add(pod5)
    #         pod2.calc_antiaffinity_pods_list_length += 4
            
    #         pod3.calc_antiaffinity_pods_list.add(pod1)
    #         pod3.calc_antiaffinity_pods_list.add(pod2)
    #         pod3.calc_antiaffinity_pods_list.add(pod4)
    #         pod3.calc_antiaffinity_pods_list.add(pod5)
    #         pod3.calc_antiaffinity_pods_list_length += 4

    #         pod4.calc_antiaffinity_pods_list.add(pod1)
    #         pod4.calc_antiaffinity_pods_list.add(pod2)
    #         pod4.calc_antiaffinity_pods_list.add(pod3)
    #         pod4.calc_antiaffinity_pods_list.add(pod5)
    #         pod4.calc_antiaffinity_pods_list_length += 4

    #         pod5.calc_antiaffinity_pods_list.add(pod1)
    #         pod5.calc_antiaffinity_pods_list.add(pod2)
    #         pod5.calc_antiaffinity_pods_list.add(pod3)
    #         pod5.calc_antiaffinity_pods_list.add(pod4)
    #         pod5.calc_antiaffinity_pods_list_length += 4
    # @planned(cost=1)
    # def mark_that_node_cant_allocate_pod_by_cpu(self,
    #     pod: Pod,
    #     node: Node,
    #     globalVar: GlobalVar):
    #     if not node in pod.nodesThatCantAllocateThisPod:
    #         assert pod.cpuRequest > node.cpuCapacity - node.currentFormalCpuConsumption
    #         pod.nodesThatCantAllocateThisPod.add(node)
    #         pod.nodesThatCantAllocateThisPod_length += 1
    #     # assert globalVar.block_policy_calculated == True
    #     globalVar.block_policy_calculated = True
    #     assert node.isNull == False
    # @planned(cost=1)
    # def mark_that_node_cant_allocate_pod_by_mem(self,
    #     pod: Pod,
    #     node: Node,
    #     globalVar: GlobalVar):
    #     assert node.isNull == False
    #     if not node in pod.nodesThatCantAllocateThisPod:
    #         assert pod.memRequest > node.memCapacity - node.currentFormalMemConsumption
    #         pod.nodesThatCantAllocateThisPod.add(node)
    #         pod.nodesThatCantAllocateThisPod_length += 1
    #     # assert globalVar.block_policy_calculated == True
    #     globalVar.block_policy_calculated = True
    # @planned(cost=1)
    # def mark_that_node_cant_allocate_pod_because_of_antiaffinity(self,
    #     target_pod: Pod,
    #     antiaffinity_pod:Pod,
    #     antiaffinity_pod_node: Node,
    #     globalVar: GlobalVar):
    #     if not antiaffinity_pod_node in target_pod.nodesThatCantAllocateThisPod:
    #         assert antiaffinity_pod in target_pod.podsMatchedByAntiaffinity
    #         # assert antiaffinity_pod == antiaffinity_pod.atNode
    #         assert antiaffinity_pod in antiaffinity_pod_node.allocatedPodList
    #         assert antiaffinity_pod_node == antiaffinity_pod.atNode
    #         target_pod.nodesThatCantAllocateThisPod.add(antiaffinity_pod_node)
    #         target_pod.nodesThatCantAllocateThisPod_length += 1
    #     # assert globalVar.block_policy_calculated == True
    #     globalVar.block_policy_calculated = True
    # @planned(cost=1)
    # def mark_that_node_cant_allocate_pod_because_of_affinity(self,
    #     target_pod: Pod,
    #     affinity_pod:Pod,
    #     node: Node,
    #     globalVar: GlobalVar):
    #     assert affinity_pod in target_pod.podsMatchedByAffinity
    #     assert node == affinity_pod.atNode
    #     if node not in target_pod.nodesThatCantAllocateThisPod and \
    #         affinity_pod not in node.allocatedPodList:
    #             target_pod.nodesThatCantAllocateThisPod.add(node)
    #             target_pod.nodesThatCantAllocateThisPod_length += 1
    #     # assert globalVar.block_policy_calculated == True
    #     globalVar.block_policy_calculated = True
    # @planned(cost=1)
    # def remove_pod_from_cluster_because_of_anitaffinity_conflict(self,
    #     target_pod: Pod,
    #     scheduler: Scheduler,
    #     globalVar: GlobalVar):
    #     # assert globalVar.block_policy_calculated == True
    #     assert target_pod.nodesThatCantAllocateThisPod_length == globalVar.amountOfNodes
    #     scheduler.podQueue.remove(target_pod)
    #     scheduler.queueLength -= 1
    #     scheduler.podQueue_excluded_pods.add(target_pod)
    #     scheduler.podQueue_excluded_pods_length += 1
    #     target_pod.target_number_of_antiaffinity_pods = 0
    #     globalVar.block_policy_calculated = True
    #     target_pod.antiaffinity_met = True
    # @planned(cost=1)
    # def remove_pod_from_antiaffinitymatched_pod_list_of_target_pod(self,
    #     target_pod: Pod,
    #     antiaffinity_pod: Pod,
    #     globalVar: GlobalVar,
    #     scheduler: Scheduler):
    #     assert antiaffinity_pod in scheduler.podQueue_excluded_pods
    #     assert antiaffinity_pod in target_pod.podsMatchedByAntiaffinity
    #     target_pod.podsMatchedByAntiaffinity.remove(antiaffinity_pod)
    #     target_pod.podsMatchedByAntiaffinity_length -= 1   
    #     target_pod.target_number_of_antiaffinity_pods -= 1
    # @planned(cost=1)
    # def remove_pod_from_cluster_because_of_affinity_conflict(self,
    #     target_pod: Pod,
    #     checked_pod:Pod,
    #     scheduler: Scheduler,
    #     globalVar: GlobalVar):
    #     assert checked_pod in target_pod.podsMatchedByAffinity
    #     assert checked_pod.calc_cantmatch_affinity == True
    #     scheduler.podQueue.remove(checked_pod)
    #     scheduler.queueLength -= 1
    #     scheduler.podQueue_excluded_pods.add(checked_pod)
    #     scheduler.podQueue_excluded_pods_length += 1
    #     target_pod.target_number_of_antiaffinity_pods = 0
    #     globalVar.block_policy_calculated = True
    #     target_pod.affinity_met = True
    @planned(cost=1)
    def mark_antiaffinity_met_because_all_antiaffinity_pods_are_matched(self,
        pod: Pod,
        globalVar: GlobalVar):
        assert pod.calc_antiaffinity_pods_list_length == pod.podsMatchedByAntiaffinity_length
        assert pod.antiaffinity_set == True
        assert globalVar.deploymentsWithAntiaffinityBalanced == True 
        # assert globalVar.block_policy_calculated == True
        pod.antiaffinity_met = True
        globalVar.block_policy_calculated = True
    @planned(cost=1)
    def mark_affinity_met_because_all_affinity_pods_are_matched(self,
        pod: Pod,
        globalVar: GlobalVar):
        assert pod.calc_affinity_pods_list_length == pod.podsMatchedByAffinity_length
        assert pod.affinity_set == True
        assert globalVar.deploymentsWithAntiaffinityBalanced == True 
        # assert globalVar.block_policy_calculated == True
        pod.affinity_met = True
        globalVar.block_policy_calculated = True
    # @planned(cost=1)
    # def mark_antiaffinity_met_because_all_antiaffinity_pods_are_matched_and_those_that_cant_dont_suite(self,
    #     pod: Pod,
    #     globalVar: GlobalVar):
    #     assert pod.calc_antiaffinity_pods_list_length == pod.target_number_of_antiaffinity_pods
    #     assert pod.antiaffinity_set == True
    #     assert globalVar.deploymentsWithAntiaffinityBalanced == True 
    #     # assert globalVar.block_policy_calculated == True
    #     pod.antiaffinity_met = True
    #     globalVar.block_policy_calculated = True
    @planned(cost=1)
    def mark_antiaffinity_met_because_antiaffinity_is_not_set(self,
        pod: Pod,
        globalVar: GlobalVar):
        assert pod.antiaffinity_set == False
        assert globalVar.deploymentsWithAntiaffinityBalanced == True 
        # assert globalVar.block_policy_calculated == True
        pod.antiaffinity_met = True
    # @planned(cost=1)
    # def mark_antiaffinity_met_because_all_antiaffinity_pods_are_matched_and_those_that_cant_dont_suite_below_the_limit_for_node_amount(self,
    #     pod: Pod,
    #     globalVar: GlobalVar):
    #     assert pod.nodesThatHaveAllocatedPodsThatHaveAntiaffinityWithThisPod_length + pod.nodesThatCantAllocateThisPod_length == globalVar.amountOfNodes_limit
    #     assert pod.antiaffinity_set == True
    #     pod.antiaffinity_met = True

class Antiaffinity_check(Antiaffinity_check_basis):       
    def generate_goal(self):
        self.generated_goal_in = []
        self.generated_goal_eq = []
        for pod1 in filter(lambda p: isinstance(p, Pod) and p.antiaffinity_set == True, self.objectList):
                self.generated_goal_eq.append([pod1.antiaffinity_met, True])
        for pod1 in filter(lambda p: isinstance(p, Pod) and p.antiaffinity_preferred_set == True, self.objectList):
                self.generated_goal_eq.append([pod1.antiaffinity_preferred_met, True])
        for pod1 in filter(lambda p: isinstance(p, Pod) and p.affinity_set == True, self.objectList):
                self.generated_goal_eq.append([pod1.affinity_met, True])
        scheduler = next(filter(lambda x: isinstance(x, Scheduler),self.objectList))
        self.generated_goal_eq.append([scheduler.status, STATUS_SCHED["Clean"]])
    def goal(self):
        for what, where in self.generated_goal_in:
            assert what in where
        for what1, what2 in self.generated_goal_eq:
            assert what1 == what2
class Antiaffinity_check_with_add_node(Antiaffinity_check):
    @planned(cost=1)
    def Add_node(self,
            node : Node,
            globalVar: GlobalVar):
        assert node.status == STATUS_NODE["New"]
        node.status = STATUS_NODE["Active"]
        globalVar.amountOfNodes += 1

class Antiaffinity_check_with_limited_number_of_pods(Antiaffinity_check_basis):
    def generate_goal(self):
        self.generated_goal_in = []
        self.generated_goal_eq = []
        for globalVar in filter(lambda p: isinstance(p, GlobalVar), self.objectList):
                self.generated_goal_eq.append([globalVar.deploymentsWithAntiaffinityBalanced, True])
        for pod1 in filter(lambda p: isinstance(p, Pod), self.objectList):
                self.generated_goal_eq.append([pod1.antiaffinity_met, True])
        scheduler = next(filter(lambda x: isinstance(x, Scheduler),self.objectList))
        self.generated_goal_eq.append([scheduler.status, STATUS_SCHED["Clean"]])

    def goal(self):
        for what, where in self.generated_goal_in:
            assert what in where
        for what1, what2 in self.generated_goal_eq:
            assert what1 == what2

    # @planned(cost=1)
    # def Set_antiaffinity_between_pods_of_deployment(self,
    #     pod1: Pod,
    #     pod2: Pod,
    #     deployment: Deployment,
    #     globalVar: GlobalVar):
    #     assert pod1.searchable == True
    #     assert pod2.searchable == True
    #     if pod1 not in pod2.podsMatchedByAntiaffinity and pod2 not in pod1.podsMatchedByAntiaffinity and pod1 != pod2:
    #         assert pod1 in deployment.podList
    #         assert pod2 in deployment.podList
    #         assert deployment.amountOfActivePods > 1
    #         assert deployment.NumberOfPodsOnSameNodeForDeployment == globalVar.maxNumberOfPodsOnSameNodeForDeployment
    #         pod1.antiaffinity_set = True
    #         pod1.podsMatchedByAntiaffinity.add(pod2)
    #         pod1.podsMatchedByAntiaffinity_length += 1
    #         pod2.antiaffinity_set = True
    #         pod2.podsMatchedByAntiaffinity.add(pod1)
    #         pod2.podsMatchedByAntiaffinity_length += 1
    #         deployment.podsMatchedByAntiaffinity_length += 2
    @planned(cost=1)
    def Set_antiaffinity_between_2_pods_of_deployment(self,
        pod1: Pod,
        pod2: Pod,
        deployment: Deployment,
        globalVar: GlobalVar):
        assert pod1.atNode == pod2.atNode
        assert pod1.searchable == True
        assert pod2.searchable == True
        assert pod1 in deployment.podList
        assert pod2 in deployment.podList
        if  pod1 not in pod2.podsMatchedByAntiaffinity and \
            pod2 not in pod1.podsMatchedByAntiaffinity and \
            pod1 != pod2:
            pod1.antiaffinity_set = True
            pod1.podsMatchedByAntiaffinity.add(pod2)
            pod1.podsMatchedByAntiaffinity_length += 1
            pod2.antiaffinity_set = True
            pod2.podsMatchedByAntiaffinity.add(pod1)
            pod2.podsMatchedByAntiaffinity_length += 1
            deployment.podsMatchedByAntiaffinity_length += 2
            # if globalVar.target_amountOfPodsWithAntiaffinity == 2 and deployment not in globalVar.DeploymentsWithAntiaffinity:
            #     globalVar.DeploymentsWithAntiaffinity.add(deployment)
            #     globalVar.DeploymentsWithAntiaffinity_length += 1
    @planned(cost=1)
    def Set_antiaffinity_between_3_pods_of_deployment(self,
        pod1: Pod,
        pod2: Pod,
        pod3: Pod,
        deployment: Deployment,
        globalVar: GlobalVar):
        assert pod1.searchable == True
        assert pod2.searchable == True
        assert pod3.searchable == True
        assert pod1 in deployment.podList
        assert pod2 in deployment.podList
        assert pod3 in deployment.podList
        assert pod1.atNode == pod2.atNode
        assert pod2.atNode == pod3.atNode
        if pod1 not in pod2.podsMatchedByAntiaffinity and \
            pod2 not in pod1.podsMatchedByAntiaffinity and \
            pod1 not in pod3.podsMatchedByAntiaffinity and \
            pod2 not in pod3.podsMatchedByAntiaffinity and \
            pod3 not in pod1.podsMatchedByAntiaffinity and \
            pod3 not in pod2.podsMatchedByAntiaffinity and \
            pod1 != pod2 and pod1 != pod3 and pod2 != pod3:
            pod1.antiaffinity_set = True
            pod1.podsMatchedByAntiaffinity.add(pod2)
            pod1.podsMatchedByAntiaffinity.add(pod3)
            pod1.podsMatchedByAntiaffinity_length += 2
            pod2.antiaffinity_set = True
            pod2.podsMatchedByAntiaffinity.add(pod1)
            pod2.podsMatchedByAntiaffinity.add(pod3)
            pod2.podsMatchedByAntiaffinity_length += 2
            pod3.antiaffinity_set = True
            pod3.podsMatchedByAntiaffinity.add(pod1)
            pod3.podsMatchedByAntiaffinity.add(pod2)
            pod3.podsMatchedByAntiaffinity_length += 2
            deployment.podsMatchedByAntiaffinity_length += 6
            # if globalVar.target_amountOfPodsWithAntiaffinity == 3 and deployment not in globalVar.DeploymentsWithAntiaffinity:
            #     globalVar.DeploymentsWithAntiaffinity.add(deployment)
            #     globalVar.DeploymentsWithAntiaffinity_length += 1
    # @planned(cost=1)
    # def Set_antiaffinity_between_4_pods_of_deployment(self,
    #     pod1: Pod,
    #     pod2: Pod,
    #     pod3: Pod,
    #     pod4: Pod,
    #     deployment: Deployment,
    #     globalVar: GlobalVar):
    #     if  pod1 not in pod2.podsMatchedByAntiaffinity and \
    #         pod1 not in pod3.podsMatchedByAntiaffinity and \
    #         pod1 not in pod4.podsMatchedByAntiaffinity and \
    #         pod2 not in pod1.podsMatchedByAntiaffinity and \
    #         pod2 not in pod3.podsMatchedByAntiaffinity and \
    #         pod2 not in pod4.podsMatchedByAntiaffinity and \
    #         pod3 not in pod1.podsMatchedByAntiaffinity and \
    #         pod3 not in pod2.podsMatchedByAntiaffinity and \
    #         pod3 not in pod4.podsMatchedByAntiaffinity and \
    #         pod1 != pod2 and pod1 != pod3 and pod2 != pod3 and pod4 != pod1 and pod4 != pod2  and pod4 != pod3:
    #         assert pod1 in deployment.podList
    #         assert pod2 in deployment.podList
    #         assert pod3 in deployment.podList
    #         assert pod4 in deployment.podList
    #         pod1.antiaffinity_set = True
    #         pod1.podsMatchedByAntiaffinity.add(pod2)
    #         pod1.podsMatchedByAntiaffinity.add(pod3)
    #         pod1.podsMatchedByAntiaffinity.add(pod4)
    #         pod1.podsMatchedByAntiaffinity_length += 3
    #         pod2.antiaffinity_set = True
    #         pod2.podsMatchedByAntiaffinity.add(pod1)
    #         pod2.podsMatchedByAntiaffinity.add(pod3)
    #         pod2.podsMatchedByAntiaffinity.add(pod4)
    #         pod2.podsMatchedByAntiaffinity_length += 3
    #         pod3.antiaffinity_set = True
    #         pod3.podsMatchedByAntiaffinity.add(pod1)
    #         pod3.podsMatchedByAntiaffinity.add(pod2)
    #         pod3.podsMatchedByAntiaffinity.add(pod4)
    #         pod3.podsMatchedByAntiaffinity_length += 3
    #         pod4.antiaffinity_set = True
    #         pod4.podsMatchedByAntiaffinity.add(pod1)
    #         pod4.podsMatchedByAntiaffinity.add(pod2)
    #         pod4.podsMatchedByAntiaffinity.add(pod3)
    #         pod4.podsMatchedByAntiaffinity_length += 3
    #         globalVar.DeploymentsWithAntiaffinity.add(deployment)
    #         globalVar.DeploymentsWithAntiaffinity_length += 1
    #         deployment.podsMatchedByAntiaffinity_length += 12
    #         if globalVar.target_amountOfPodsWithAntiaffinity == 4 and deployment not in globalVar.DeploymentsWithAntiaffinity:
    #             globalVar.DeploymentsWithAntiaffinity.add(deployment)
    #             globalVar.DeploymentsWithAntiaffinity_length += 1
    @planned(cost=1)
    def Reduce_maxNumberOfPodsOnSameNodeForDeployment(self,
        globalVar: GlobalVar):
        globalVar.maxNumberOfPodsOnSameNodeForDeployment -= 1       
   
    @planned(cost=1)
    def Calculate_fulfilled_antiaffinity_pods_of_deployment(self,
        pod: Pod,
        deployment: Deployment,
        permutations_of_amountOfActivePods: Permutations,
        globalVar: GlobalVar):
        if deployment not in globalVar.DeploymentsWithAntiaffinity: 
            assert deployment.amountOfActivePods > 1
            assert permutations_of_amountOfActivePods.number == deployment.amountOfActivePods
            assert deployment.podsMatchedByAntiaffinity_length >= permutations_of_amountOfActivePods.permutations

            # if deployment.amountOfActivePods > 10:
            #     assert pod.podsMatchedByAntiaffinity_length == 45 
            # else:    
            #     assert permutations_of_amountOfActivePods.number == deployment.amountOfActivePods
            #     assert pod.podsMatchedByAntiaffinity_length == permutations_of_amountOfActivePods.permutations - 1
            globalVar.DeploymentsWithAntiaffinity.add(deployment)
            globalVar.DeploymentsWithAntiaffinity_length += 1

    @planned(cost=1)
    def Calculate_fulfilled_antiaffinity_pods_of_deployment_with_pod_limit(self,
        deployment: Deployment,
        permutations_of_target_amountOfPodsWithAntiaffinity: Permutations,
        globalVar: GlobalVar):
        if deployment not in globalVar.DeploymentsWithAntiaffinity: 
            assert deployment.amountOfActivePods > 1
            assert globalVar.target_amountOfPodsWithAntiaffinity > 0
            assert permutations_of_target_amountOfPodsWithAntiaffinity.number == globalVar.target_amountOfPodsWithAntiaffinity
            assert deployment.podsMatchedByAntiaffinity_length >= permutations_of_target_amountOfPodsWithAntiaffinity.permutations
            # if globalVar.target_amountOfPodsWithAntiaffinity > 10:
            #     assert pod.podsMatchedByAntiaffinity_length == 45 
            # else:    
            #     assert permutations_of_target_amountOfPodsWithAntiaffinity.number == globalVar.target_amountOfPodsWithAntiaffinity
            #     assert pod.podsMatchedByAntiaffinity_length == permutations_of_target_amountOfPodsWithAntiaffinity.permutations - 1
            globalVar.DeploymentsWithAntiaffinity.add(deployment)
            globalVar.DeploymentsWithAntiaffinity_length += 1

    @planned(cost=1)
    def mark_that_antiaffinity_is_set_for_requested_number_of_deployments(self, 
        globalVar: GlobalVar):
        assert globalVar.DeploymentsWithAntiaffinity_length == globalVar.target_DeploymentsWithAntiaffinity_length
        globalVar.deploymentsWithAntiaffinityBalanced = True
    
class Antiaffinity_check_with_limited_number_of_pods_with_add_node(Antiaffinity_check_with_limited_number_of_pods):
    @planned(cost=1)
    def Add_node(self,
            node : Node,
            globalVar: GlobalVar):
        assert node.status == STATUS_NODE["New"]
        node.status = STATUS_NODE["Active"]
        globalVar.amountOfNodes += 1

class Balance_pods_and_drain_node(Antiaffinity_check_with_limited_number_of_pods):
    @planned(cost=1)
    def DrainNode(self,
        node: "Node",
        globalVar: GlobalVar):
        assert node.amountOfActivePods == 0
        assert node.status == STATUS_NODE["Active"]
        assert node.isNull == False
        globalVar.NodesDrained_length += 1
        globalVar.NodesDrained.add(node)
        node.status = STATUS_NODE["Inactive"]

        self.script.append(script_remove_node(node))
    
    @planned(cost=1)
    def Reqested_amount_of_nodes_drained(self,
        globalVar: GlobalVar):
        assert globalVar.NodesDrained_length == globalVar.target_NodesDrained_length
        globalVar.target_NodesDrained_length_reached = True

    def generate_goal(self):
        self.generated_goal_in = []
        self.generated_goal_eq = []
        for globalVar in filter(lambda p: isinstance(p, GlobalVar), self.objectList):
                self.generated_goal_eq.append([globalVar.deploymentsWithAntiaffinityBalanced, True])
        for pod1 in filter(lambda p: isinstance(p, Pod), self.objectList):
                self.generated_goal_eq.append([pod1.antiaffinity_met, True])
        for globalVar in filter(lambda p: isinstance(p, GlobalVar), self.objectList):
            self.generated_goal_eq.append([globalVar.target_NodesDrained_length_reached, True])
        scheduler = next(filter(lambda x: isinstance(x, Scheduler),self.objectList))
        self.generated_goal_eq.append([scheduler.status, STATUS_SCHED["Clean"]])

    # def goal(self):
    #     for what, where in self.generated_goal_in:
    #         assert what in where
    #     for what1, what2 in self.generated_goal_eq:
    #         assert what1 == what2

    



import subprocess
import json
from kalc.model.kinds.Node import Node
import kalc.model.kinds.ReplicaSet as rs
import kalc.model.kinds.Deployment as d
import time

def script_remove_node(node):
    if hasattr(node, "_variable_mode") and node._variable_mode: 
        return "" # FIX for https://github.com/criticalhop/poodle/issues/59
    return f'echo "Please remove node \'{str(node.metadata_name)}\'"' 


def get_rs_from_deployment(deployment, object_space):
    replicasets = filter(lambda x: isinstance(x, rs.ReplicaSet), object_space)
    for replicaset in replicasets:
        if replicaset.metadata_ownerReferences__name == deployment.metadata_name:
            return replicaset
    raise ValueError("Can not find ReplicaSet")


def get_deployment_from_pod(pod, object_space):
    deployments = filter(lambda x: isinstance(x, d.Deployment), object_space)
    for deployment in deployments:
        if pod in deployment.podList:
            return deployment 
    return None


def move_pod_with_deployment_script_simple(pod, node_to: Node, object_space):
    if hasattr(pod, "_variable_mode") and pod._variable_mode: 
        return "" # FIX for https://github.com/criticalhop/poodle/issues/59
    d = get_deployment_from_pod(pod, object_space)
    return move_pod_with_deployment_script(pod, node_to, d,
        get_rs_from_deployment(d, object_space))


def move_pod_with_deployment_script(pod, node_to: Node, deployment, replicaset):
    "Move the pod when the pod is part of Deployment"
    # 1. Dump full original Deployment
    deployment_name = str(deployment.metadata_name)
    deployment_namespace = deployment.metadata_namespace._get_value()
    replicaset_name = str(replicaset.metadata_name)
    replicaset_namespace =  replicaset.metadata_namespace._get_value()
    pod_original_name = str(pod.metadata_name)
    pod_namespace =  pod.metadata_namespace._get_value()
    pod_new_name = f"{pod_original_name}-kalcmoved-" + str(time.time())

    assert str(pod_namespace) == str(deployment_namespace), "{0}({1}) == {2}({3})".format(pod_namespace, pod_original_name, deployment_namespace, deployment_name)
    assert str(pod_namespace) == str(replicaset_namespace), "{0}({1}) == {2}({3})".format(pod_namespace, pod_original_name, replicaset_namespace, replicaset_name)
    namespace = pod_namespace

    if namespace is None:
        namespace = 'default'
    
    # TODO: check that pod that we are deleting had a full green-light status (alive&ready)
    # TODO: check if state has diverged too far and we can not continue
    # TODO: explicily prohibit moving singleton pods... except allowed explicitly
    # TODO: namespace support

    # TODO: dry run mode!! - default mode
    # TODO: move fake pod - test mode
    # TODO: move smth unimportant
    # TODO: execute mode
    # TODO: paranoid checks: that deployment does not get re-created
    
    from kalc.interactive import cluster_md5_sh, md5_cluster

    move_pod = f"""

echo "Moving pod '{pod_original_name}'..."
echo " -- Disabling relevant controllers by backing up and temporarily deleting them..."
kubectl get deployment/{deployment_name} --namespace={namespace} -o=yaml > ./deployment_{deployment_name}.yaml &&
kubectl delete --cascade=false deployment/{deployment_name} --namespace={namespace} &&
kubectl get replicaset/{replicaset_name} --namespace={namespace} -o=yaml > ./replicaset_{replicaset_name}.yaml &&
kubectl delete --cascade=false replicaset/{replicaset_name} --namespace={namespace} &&
echo " -- Storing current version of pod config of the pod-to-be-moved..." &&
kubectl get pod/{pod_original_name} --namespace={namespace} -o=yaml > ./pod_new.yaml &&
echo "  --- Renaming pod template..." &&
yq '.metadata += {{name: "{pod_new_name}"}}' ./pod_new.yaml > ./pod_new.yaml.$$ && mv ./pod_new.yaml.$$ pod_new.yaml &&
echo "  --- Deleting status from dump..." &&
yq 'del(.status)' ./pod_new.yaml > ./pod_new.yaml.$$ && mv ./pod_new.yaml.$$ pod_new.yaml &&
echo "  --- Inserting nodeName..." &&
yq '.spec += {{nodeName: "{str(node_to.metadata_name)}"}}' ./pod_new.yaml > ./pod_new.yaml.$$ && mv ./pod_new.yaml.$$ pod_new.yaml &&
echo " -- Running new pod..." &&
kubectl apply -f ./pod_new.yaml &&
echo " -- Waiting 30s for new pod to become ready..." &&
kubectl wait --for condition=ready --namespace={namespace} -f ./pod_new.yaml &&
echo " -- Deleting original pod..." &&
kubectl delete pod/{pod_original_name} --namespace={namespace}
echo " -- Re-applying ReplicaSet..." &&
kubectl apply -f ./replicaset_{replicaset_name}.yaml &&
echo " -- Re-applying Deployment..." &&
kubectl apply -f ./deployment_{deployment_name}.yaml &&
echo "Done moving pod '{pod_original_name}'!" """
    return move_pod

def generate_compat_header():
    "Generate script header for checking if correct tools are installed"
    
    from kalc.interactive import cluster_md5_sh, md5_cluster

    # Compatibility and installed utilities part

    compat = f"""#!/bin/bash
die() {{ echo "$*" 1>&2 ; exit 1; }}

# Checking for tools
echo "Checking for kubectl..." && kubectl > /dev/null || die "sed not found" 
echo "Checking for jq..." && jq --version | grep -q jq- || die "jq not found or not compatible. Install with 'apt install jq'" 
echo "Checking for yq..." && yq --version 2>&1 | grep -q "yq 2" || die "yq not found or not compatible" 
echo "Checking for sed..." && sed --version >/dev/null 2> /dev/null || die "sed not found"
echo "Checking for awk..." && awk --version >/dev/null 2> /dev/null || die "awk not found"
echo -n "Checking for kubectl get permission for deployment... " && kubectl auth can-i get deployment || die "kubectl does not have permission to get deployments" 
echo -n "Checking for kubectl get permission for replicaset... " && kubectl auth can-i get replicaset || die "kubectl does not have permission to get replicaset" 
echo -n "Checking for kubectl get permission for pod... " && kubectl auth can-i get pod || die "kubectl does not have permission to get pod" 
echo -n "Checking for kubectl apply permission for deployment... " && kubectl auth can-i apply deployment || die "kubectl does not have permission to apply deployments" 
echo -n "Checking for kubectl apply permission for replicaset... " && kubectl auth can-i apply replicaset || die "kubectl does not have permission to apply replicaset" 
echo -n "Checking for kubectl apply permission for pod... " && kubectl auth can-i apply pod || (echo "kubectl does not have permission to apply pod" && exit 1)

CLUSTER_MD5={md5_cluster}
CLUSTER_MD5_current=`{cluster_md5_sh} | awk ' {{printf $1}} ' `
if [ "$CLUSTER_MD5" !=  "$CLUSTER_MD5_current" ] ;
then
 echo "WARNING: CLUSTER STATE DIVERGED - USE AT YOUR OWN RISK"
  finish="1"
  if [ "$1" != "-y" ];
  then
    echo  -n "Are you going to continue?  [yes/no]? "
    read answer
    finish="-1"
  fi
  while [ "$finish" = '-1' ]
  do
    finish="1"
    if [ "$answer" = '' ];
    then
      answer=""
    else
      case $answer in
        yes | YES ) answer="y";;
        no | NO ) exit 0 ;;
        *) finish="-1";
           echo -n 'Invalid !!! Please enter "yes" or "no" :';
           read answer;;
       esac
    fi
  done
fi
"""
    return compat


from poodle import * 
from kalc.model.system.base import ModularKind
from kalc.model.system.Controller import Controller
from kalc.model.system.base import HasLimitsRequests
from kalc.model.kinds.Node import Node
import kalc.model.kinds.Pod as mpod
from kalc.model.system.primitives import Status


class LoadBalancer(ModularKind):
    def __init__(self, *args, **kwargs):
        super().__init__( *args, **kwargs)
    


import random
from typing import Set
from kalc.model.system.primitives import Label, StatusServ
import kalc.model.kinds.Pod as mpod
from kalc.model.system.base import HasLabel, ModularKind
import kalc.model.kinds.Pod as mpod
from kalc.model.system.primitives import StatusSched
from kalc.misc.const import *


class Service(ModularKind, HasLabel):
    spec_selector: Set[Label]
    amountOfActivePods: int
    amountOfPodsInQueue: int
    amountOfPodsOnDifferentNodes: int
    targetAmountOfPodsOnDifferentNodes: int
    status: StatusServ
    metadata_name: str
    metadata_namespace: str
    searchable: bool
    isNull: bool
    isSearched: bool
    podList: Set["mpod.Pod"]
    antiaffinity: bool
    policy_antiaffinity_prefered : bool
    antiaffinity_prefered_policy_met: bool
    
    
    def __init__(self, *args, **kwargs):
        super().__init__( *args, **kwargs)
        self.metadata_name = "modelService"+str(random.randint(100000000, 999999999))
        # self.metadata_name = "model-default-name"
        self.amountOfActivePods = 0
        self.status = STATUS_SERV["Pending"]
        self.searchable = True
        self.isNull = False
        self.isSearched = False
        self.amountOfPodsInQueue = 0
        self.amountOfPodsOnDifferentNodes = 0
        self.targetAmountOfPodsOnDifferentNodes = 0
        self.policy_antiaffinity_prefered = False    
    def __str__(self): return str(self.metadata_name)

    # def __repr__(self):
    #     return 'Servicename : ' + str(self._get_value()) 

Service.SERVICE_NULL = Service("NULL")
Service.SERVICE_NULL.metadata_name = "Null-Service"
Service.SERVICE_NULL.isNull = True
Service.SERVICE_NULL.searchable == False





from kalc.interactive import *
from kalc.model.search import Balance_pods_and_drain_node
from kalc.model.kinds.Deployment import Deployment
from kalc.misc.script_generator import generate_compat_header
from collections import defaultdict
from poodle.schedule import SchedulingError
from itertools import combinations, product

D_RANK = 0
D_DEPLOYMENT = 1
D_PODS = 2
D_UNBALANCED_PODS = 3

L_TARGETS = 0
L_DEPLOYMENTS = 1
L_NODES = 2
L_PODS = 3


def generate_hypothesys_combination(deployments, nodes):
    deployments_maxpods = []
    drain_node_frequency = 2
    twice_drain_node_frequency = 3
    deployment_amount = 0
    searchable_deployments = set()
    searchable_pods = set()
    deployments_by_ranks = []
    max_pod_number = 0
    comb_nodes_pods_fitered = []
    for d in deployments:
        deployment_amount += 1
        deployment_max_pod_number = 0
        list_for_comb = []
        pod_node = defaultdict(list)
        if len(d.podList._get_value()) == 0:
            continue
        for pod in d.podList:
            pods_node = pod.atNode._property_value # Poodle bug. https://github.com/criticalhop/poodle/issues/122
            pod_node[str(pods_node.metadata_name)].append(pod) # str: poodle bug
            deployment_max_pod_number += 1
        # Find which pods are the worst
        unbalanced_pods_indexed = [[len(x), x] for x in pod_node.values()]
        unbalanced_pods_indexed.sort(key=lambda x: x[0], reverse=True)
        unbalanced_pods = [x[1] for x in unbalanced_pods_indexed]
        only_unbalanced = unbalanced_pods_indexed[0][1]
        unbalanced_pods = [item for sublist in unbalanced_pods for item in sublist] # flatten
        if deployment_max_pod_number > max_pod_number:
            max_pod_number = deployment_max_pod_number
        deployments_maxpods.append([max([x[0] for x in unbalanced_pods_indexed]), d, unbalanced_pods, only_unbalanced])
    deployments_maxpods = sorted(deployments_maxpods, key=lambda x: x[0], reverse=True) 
    highest_max_pods_deployment = max(x[0] for x in deployments_maxpods)
    list_of_deployments_sorted = []
    prev_deployments_current_rank=[]
    list_deployments_targets = []
    for processed_rank in range(highest_max_pods_deployment,0,-1):
        deployments_current_rank = []
        for d in deployments_maxpods:
            if d[0] == processed_rank:
                deployments_current_rank.append(d[1])
            if processed_rank == 0:
                deployments_current_rank.append(d[1])
        if deployments_current_rank:
            deployments_current_rank.extend(prev_deployments_current_rank)
            list_of_deployments_sorted.append(deployments_current_rank)
            prev_deployments_current_rank = deployments_current_rank
    print(f"Worst case deployment {str(deployments_maxpods[0][D_DEPLOYMENT])}, with {deployments_maxpods[0][D_RANK]} pods on same node")
    list_deployments_targets = list(range(1,deployment_amount+1))
    list_nodes = list(range(0,2))
    list_pods = list(range(2,max_pod_number+1))

    list_for_comb.append(list_deployments_targets)
    list_for_comb.append(list_of_deployments_sorted)
    list_for_comb.append(list_nodes)
    list_for_comb.append(list_pods)
    comb_nodes_pods = list(product(*list_for_comb))
    #TODO: Exclude combinations when number of serachable deployments is less than list of deployments 
    comb_nodes_pods_fitered = []
    for comb in comb_nodes_pods:
        if len(comb[1]) >= comb[0]:
            comb_nodes_pods_fitered.append(comb)
    return comb_nodes_pods_fitered


def optimize_cluster(clusterData=None):
    print("WARNING! Not taking into account service SLOs")
    update(clusterData) # To reload from scratch...
    deployments = list(filter(lambda x: isinstance(x, Deployment), kalc_state_objects)) # to get amount of deployments
    nodes = list(filter(lambda x: isinstance(x, Node), kalc_state_objects))
    comb_nodes_pods = generate_hypothesys_combination(deployments,nodes)
    for combination in comb_nodes_pods:
        update(clusterData) # To reload from scratch...
        problem = Balance_pods_and_drain_node(kalc_state_objects)
        deployments_local = list(filter(lambda x: isinstance(x, Deployment), kalc_state_objects))
        globalVar_local = next(filter(lambda x: isinstance(x, GlobalVar), kalc_state_objects))
        pods_local = list(filter(lambda x: isinstance(x, Pod), kalc_state_objects))
        d_cand = []
        p_cand = []
        for d in deployments_local:
            combination_metadatanames = []
            for c in  combination[L_DEPLOYMENTS]:
                combination_metadatanames.append(c.metadata_name)
            if str(d.metadata_name) in  combination_metadatanames:
                d.searchable = True
                d_cand.append(str(d.metadata_name))
                for spod in pods_local:
                    if spod in d.podList:
                        spod.searchable = True
                        p_cand.append(str(spod.metadata_name))
        globalVar_local.target_DeploymentsWithAntiaffinity_length = combination[L_TARGETS]
        globalVar_local.target_amountOfPodsWithAntiaffinity = combination[L_PODS]
        globalVar_local.target_NodesDrained_length = combination[L_NODES]

        print("Deployment candidates:", ', '.join(d_cand))
        print("Pod candidates:", ", ".join(p_cand))
        print("-----------------------------------------------------------------------------------")
        print("--- Solving case for deployment_amount =", combination[L_TARGETS], ", pod_amount =", combination[L_PODS], ", drain nodes = ", combination[L_NODES], " ---")
        print("-----------------------------------------------------------------------------------")
        try:
            problem.xrun()
        except SchedulingError:
            print("Could not solve in this configuration, trying next...")
            continue
        move_script = '\n'.join(problem.script)
        full_script = generate_compat_header() + move_script
        scritpt_file = f"./kalc_optimize_{combination[L_TARGETS]}_{combination[L_PODS]}_{combination[L_NODES]}.sh"
        print("Generated optimization script at", scritpt_file)
        with open(scritpt_file, "w+") as fd:
            fd.write(full_script)
            
            
def run():
    optimize_cluster(None)



from poodle import schedule, xschedule
from poodle.schedule import SchedulingError
from kalc.model.system.math import permutation_list

class ProblemTemplate:
    def __init__(self, objectList=None):
        if objectList is None: self.objectList = []
        else: self.objectList = objectList
        self.objectList.extend(permutation_list)
        self.plan = None
        self.pod = []
        self.node = []
        self.kubeProxy = []
        self.loadbalancer = []
        self.service = []
        self.controller = []
        self.request = []
        self.containerConfig = []
        self.priorityDict = {}
        self.goals_in = []
        self.goals_eq = []
        self.external_actions = []
        self.lambda_goal = []
        self.script = []
    
    def problem(self):
        pass
        
    def addObject(self, obj):
        self.objectList.append(obj)
        return obj

    # fill object with corresponding list (list instance in lower case)
    def fillObjectLists(self):
        requiredObject = ['Pod', "Node", "Service", "LoadBalancer", "DaemonSet", "Deployment"]
        for obj in self.objectList:
            if obj.__class__.__name__ in requiredObject:
                try:
                    getattr(self, obj.__class__.__name__.lower()).append(obj)
                except:
                    pass
            
    def generate_goal(self):
        pass

    def add_external_method(self, action):
        if not action in self.external_actions: self.external_actions.append(action)

    def run(self, timeout=999000, sessionName=None, schedule=schedule, raise_=False):
        if not sessionName: sessionName = self.__class__.__name__
        self.problem()
        self.generate_goal()
        self_methods = [getattr(self,m) for m in dir(self) if callable(getattr(self,m)) and hasattr(getattr(self, m), "_planned")]
        model_methods = []
        methods_scanned = set()
        for obj in self.objectList:
            if not obj.__class__.__name__ in methods_scanned:
                methods_scanned.add(obj.__class__.__name__)
                for m in dir(obj):
                    if callable(getattr(obj, m)) and hasattr(getattr(obj, m), "_planned"):
                        model_methods.append(getattr(obj, m))
        try:
            self.plan = schedule(
                methods=self_methods + list(model_methods) + self.external_actions, 
                # methods=self_methods + list(model_methods), 
                space=list(self.__dict__.values())+self.objectList,
                goal=lambda:(self.goal()),
                timeout=timeout,
                sessionName=sessionName
                #exit=self.exit
            )
        except SchedulingError:
            if raise_:
                raise SchedulingError("Can't solve")
            else:
                pass
    
    def xrun(self, timeout=999000, sessionName=None):
        "Run and execute plan"
        self.run(timeout, sessionName, schedule=xschedule, raise_=True)

    def goal(self):
        pass

    


from kalc.model.system.base import ModularKind
from kalc.model.system.Controller import Controller
from kalc.model.system.base import HasLimitsRequests
from kalc.model.kinds.PriorityClass import PriorityClass, zeroPriorityClass
from kalc.model.system.Scheduler import Scheduler
import kalc.model.kinds.Pod as mpod
import kalc.model.system.globals as mGlobalVar
from kalc.model.kinds.ReplicaSet import ReplicaSet
from kalc.model.system.primitives import Status, Label
from kalc.misc.const import STATUS_POD, STATUS_SCHED, StatusDeployment
import kalc.model.kinds.Node as mnode
import kalc.model.kinds.Deployment as mdeployment
from poodle import *
from typing import Set
from logzero import logger
import kalc.misc.util as util
import random
import yaml, copy, jsonpatch, difflib

def calculate_maxNumberOfPodsOnSameNode_metrics(self, object_space):
    deployments = filter(lambda x: isinstance(x, mdeployment.Deployment), object_space)
    pods = filter(lambda x: isinstance(x, mpod.Pod), object_space)
    nodes = filter(lambda x: isinstance(x,mnode.Node), object_space)
    globalVar = next(filter(lambda x: isinstance(x, mGlobalVar.GlobalVar), object_space))
    deploymentController_max_node_amount_of_pods_list = []
    for deploymentController in deployments:
        node_amount_of_pods_list = []
        for node in nodes:
            amount_of_deployment_pod_on_node = 0
            for pod in node.allocatedPodList:
                if pod in deploymentController.podList:
                    amount_of_deployment_pod_on_node += 1
            node_amount_of_pods_list.append(amount_of_deployment_pod_on_node)
        if len(node_amount_of_pods_list) == 0:
            deploymentController.NumberOfPodsOnSameNodeForDeployment = 1
        else:
            deploymentController.NumberOfPodsOnSameNodeForDeployment = max(node_amount_of_pods_list)
        deploymentController_max_node_amount_of_pods_list.append(deploymentController.NumberOfPodsOnSameNodeForDeployment)
    if len(deploymentController_max_node_amount_of_pods_list) == 0:
        globalVar.maxNumberOfPodsOnSameNodeForDeployment = 1
    else:
        globalVar.maxNumberOfPodsOnSameNodeForDeployment = max(deploymentController_max_node_amount_of_pods_list)


import random
from kalc.model.system.base import ModularKind
from kalc.model.system.Controller import Controller
from kalc.model.system.base import HasLimitsRequests
from kalc.model.kinds.Node import Node
from kalc.model.kinds.PriorityClass import PriorityClass, zeroPriorityClass
from kalc.model.system.Scheduler import Scheduler
import kalc.model.kinds.Pod as mpod
from kalc.model.system.primitives import StatusDaemonSet
from kalc.misc.const import *
from poodle import *
from typing import Set
from logzero import logger
import kalc.misc.util as util

#TODO fill pod-template-hash with https://github.com/kubernetes/kubernetes/blob/0541d0bb79537431421774465721f33fd3b053bc/pkg/controller/controller_utils.go#L1024
class DaemonSet(ModularKind, Controller, HasLimitsRequests):
    metadata_name: str
    metadata_uid: str
    metadata_namespace: str
    amountOfActivePods: int
    status: StatusDaemonSet
    podList: Set["mpod.Pod"]
    spec_template_spec_priorityClassName: str

    searchable: bool


    def __init__(self, *args, **kwargs):
        super().__init__( *args, **kwargs)
        self.metadata_name = "modelDaemonSet"+str(random.randint(100000000, 999999999))
        # self.metadata_name = "model-default-name"
        self.amountOfActivePods = 0
        self.searchable = True
        self.spec_template_spec_priorityClassName = "Normal-zero"
        self.priorityClass = zeroPriorityClass
        self.status = STATUS_DAEMONSET_PENDING
        self.metadata_uid = "undefined"
        #TODO make support for DAEMONSET start status while loading

    def hook_after_create(self, object_space):
        # TODO throw error if name already exist
        # Error from server (AlreadyExists): error when creating "./tests/daemonset_eviction/daemonset_create.yaml": daemonsets.apps "fluentd-elasticsearch" already exists
        nodes = filter(lambda x: isinstance(x, Node), object_space)
        i = 0
        for node in nodes:
            i += 1
            new_pod = mpod.Pod()
            new_pod.metadata_name = str(self.metadata_name) + '-DaemonSet_CR-' + str(i)
            new_pod.toNode = node
            new_pod.cpuRequest = self.cpuRequest
            new_pod.memRequest = self.memRequest
            new_pod.cpuLimit = self.cpuLimit
            new_pod.memLimit = self.memLimit
            new_pod.status = STATUS_POD["Pending"]
            new_pod.hook_after_load(object_space, _ignore_orphan=True) # for service<>pod link, add to sched
            new_pod.set_priority(object_space, self)
            new_pod.hasDaemonset = True
            self.podList.add(new_pod)
            object_space.append(new_pod)
            self.amountOfActivePods += 1

    def hook_after_load(self, object_space):
        daemonSets = filter(lambda x: isinstance(x, DaemonSet), object_space)
        for daemonSetController in daemonSets:
            if daemonSetController != self and str(daemonSetController.metadata_name) == str(self.metadata_name):
                message = "Error from server (AlreadyExists): DaemonSet with name {0}".format(self.metadata_name)
                logger.error(message)
                raise AssertionError(message)
        pods = filter(lambda x: isinstance(x, mpod.Pod), object_space)
        nodes = filter(lambda x: isinstance(x, Node), object_space)

        for pod in pods:
            if str(pod.metadata_ownerReferences__name) == str(self.metadata_name):
                self.podList.add(pod)
                self.amountOfActivePods += 1
                pod.hasDaemonset = True
                for node in nodes:
                    if node.metadata_name._get_value() == pod.spec_nodeName._get_value():
                        pod.toNode = node
                # self.check_pod(pod, object_space)

    def hook_after_apply(self, object_space):
        daemonSets = filter(lambda x: isinstance(x, DaemonSet), object_space)
        old_daemonSet = self
        for daemonSetController in daemonSets:
            if daemonSetController != self and str(daemonSetController.metadata_name) == str(self.metadata_name):
                old_daemonSet = daemonSetController
                break
        # if old DEployment not found
        if old_daemonSet == self:
            self.hook_after_create(object_space)
        else:
            self.podList = old_daemonSet.podList # copy pods instead of creating
            #scale memory and cpu
            for pod in util.objDeduplicatorByName(self.podList._get_value()):
                pod.cpuRequest = self.cpuRequest
                pod.memRequest = self.memRequest
                pod.cpuLimit = self.cpuLimit
                pod.memLimit = self.memLimit
                pod.set_priority(object_space, self)
            object_space.remove(old_daemonSet) # delete old Deployment
    def __str__(self): return str(self._get_value())





from poodle import Object

 
class Type(Object):
    def __str__(self): return str(self._get_value())

class Status(Object):
    def __str__(self): return str(self._get_value())

class StatusPod(Object):
    def __str__(self): return str(self._get_value())

class StatusNode(Object):
    def __str__(self): return str(self._get_value())

class StatusReq(Object):
    def __str__(self): return str(self._get_value())

class StatusSched(Object):
    def __str__(self): return str(self._get_value())
    
class StatusServ(Object):
    def __str__(self): return str(self._get_value())

class StatusDeployment(Object):
    def __str__(self): return str(self._get_value())

class StatusDaemonSet(Object):
    def __str__(self): return str(self._get_value())

class StatusLim(Object):
    pass

class Label(Object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "key" in kwargs and "value" in kwargs:
            self.key = kwargs["key"]
            self.value = kwargs["value"]
        elif len(args) and isinstance(args[0], str) and ":" in args[0]:
            self.key, self.value = args[0].split(":")
        else:
            self.key="<UNDEFINED>"
            self.value="<UNDEFINED>"
    def __str__(self):
        return str(self._get_value())

class TypePolicy(Object):
    pass

class TypeServ(Object):
    pass


import sys
from poodle import planned
from logzero import logger
from kalc.model.system.base import HasLimitsRequests, HasLabel
from kalc.model.system.primitives import TypeServ
from kalc.model.system.Controller import Controller
from kalc.model.system.primitives import Label
from kalc.model.kinds.Service import Service
from kalc.model.kinds.Deployment import Deployment
from kalc.model.kinds.DaemonSet import DaemonSet
from kalc.model.kinds.Pod import Pod
from kalc.model.kinds.Node import Node
from kalc.model.kinds.PriorityClass import PriorityClass, zeroPriorityClass
from kalc.misc.const import *
from kalc.misc.problem import ProblemTemplate
from kalc.misc.util import cpuConvertToAbstractProblem, memConvertToAbstractProblem

class LabelsModel(ProblemTemplate):
    def add_goal_in(self, goal_entry):
        self.goals_in.extend(goal_entry)

    def add_goal_eq(self, goal_entry):
        self.goals_eq.extend(goal_entry)

    def generate_goal(self):
        self.add_goal_in([[self.pod1, pod2.podsMatchedByAntiaffinity]])

    def goal(self):
        if self.goals_in:
            for what, where in self.goals_in:
                assert what in where
        if self.goals_eq:
            for what1, what2 in self.goals_eq:
                assert what1 == what2
    def problem(self):
        self.scheduler = next(filter(lambda x: isinstance(x, Scheduler), self.objectList))
        self.globalVar = next(filter(lambda x: isinstance(x, GlobalVar), self.objectList))

    @planned(cost=1)
    def add_label_to_pod(self,
        pod: Pod,
        label: Label):
        pod.metadata_labels.add(label)

    @planned(cost=1)
    def add_antiaffinity_label_to_pod(self,
        pod: Pod,
        label: Label):
        pod.antiaffinity_labels.add(label)
        pod.antiaffinity_labels_length += 1
    @planned(cost=1)
    def remove_label_from_pod(self,
        pod: Pod,
        label: Label):
        pod.metadata_labels.add(label)

    @planned(cost=1)
    def remove_antiaffinity_label_from_pod(self,
        pod: Pod,
        label: Label):
        pod.antiaffinity_labels.add(label)
        pod.antiaffinity_labels_length += 1
    
    @planned(cost=1)
    def add_label_to_label_connection_with_antiaffinity(self,
        pod_source: Pod,
        pod_target: Pod,
        label: Label,
        label_connection: Label_connection):
        assert pod_source.is_checked_as_source_for_labels == True
        assert pod_target.is_checked_as_target_for_labels == True
        assert pod_source == label_connection.pod_source
        assert pod_target == label_connection.pod_target
        assert label in pod_source.antiaffinity_labels
        assert label in pod_target.metadata_labels
        assert label in label_connection.matching_labels_candidates        

        label_connection.matching_labels_candidates.remove(label)
        label_connection.Label_connection.add(label)
        label_connection.Label_connection_length += 1

    @planned(cost=1)
    def Mark_pod_as_matching(self,
                service1: Service,
                scheduler: "Scheduler",
                globalVar: GlobalVar,
                pod_source: Pod,
                pod_target: Pod,
                label_connection: Label_connection
         ):
        assert pod_source.is_checked_as_source_for_labels == True
        assert pod_target.is_checked_as_target_for_labels == True
        assert pod_source == label_connection.pod_source
        assert pod_target == label_connection.pod_target
        assert label_connection.Label_connection_length == pod_source.antiaffinity_labels_length
        pod_source.antiaffinity_pods_list.add(pod_target)

        


import subprocess
import json
from poodle import Object
from kalc.model.full import kinds_collection
from kalc.misc.kind_filter import FilterByLabelKey, FilterByName, KindPlaceholder
from kalc.model.kubernetes import KubernetesCluster
import kalc.policy 
from kalc.model.search import KubernetesModel
from kalc.model.kinds.Deployment import YAMLable, Deployment
from pygments import highlight
from pygments.lexers.diff import DiffLexer
from pygments.formatters.terminal import TerminalFormatter
import random
import io
import kalc.misc.util
import pkg_resources
__version__ = pkg_resources.get_distribution("kalc").version

cluster_md5_sh = 'kubectl get pods -o wide --all-namespaces -o=custom-columns=NAME:.metadata.name,NODE:.spec.nodeName --sort-by="{.metadata.name}" | md5sum'

kalc_state_objects = []
kind = KindPlaceholder
cluster = None

md5_cluster = ""

kalc.policy.policy_engine.register_state_objects(kalc_state_objects)

for k, v in kinds_collection.items():
    v.by_name = FilterByName(k, kalc_state_objects)
    v.by_label = FilterByLabelKey(k, kalc_state_objects)
    globals()[k] = v
    setattr(kind, k, v)

def update(data=None):
    "Fetch information from currently selected cluster"
    if isinstance(data, io.IOBase):
        data = data.read()
    k = KubernetesCluster()
    if not data:
        global md5_cluster
        result = subprocess.Popen(cluster_md5_sh, shell=True, stdout=subprocess.PIPE, executable='/bin/bash')
        md5_cluster = result.stdout.read().decode('ascii').split()[0]
        assert len(md5_cluster) == 32, "md5_cluster sum wrong len({0}) not is 32".format(md5_cluster)

        result = subprocess.run(['kubectl', 'get', 'all', '--all-namespaces', '-o=json'], stdout=subprocess.PIPE)
        if len(result.stdout) < 100:
            print(result.stdout)
            raise SystemError("Error using kubectl. Make sure `kubectl get pods` is working.")
        data = json.loads(result.stdout.decode("utf-8"))
        for item in data["items"]:
            k.load_item(item)

        result = subprocess.run(['kubectl', 'get', 'node', '-o=json'], stdout=subprocess.PIPE)
        if len(result.stdout) < 100:
            raise SystemError("Error using kubectl. Make sure `kubectl get pods` is working.")
        data = json.loads(result.stdout.decode("utf-8"))
        for item in data["items"]:
            k.load_item(item)
        result = subprocess.run(['kubectl', 'get', 'pc', '-o=json'], stdout=subprocess.PIPE)
        if len(result.stdout) < 100:
            raise SystemError("Error using kubectl. Make sure `kubectl get pods` is working.")
        data = json.loads(result.stdout.decode("utf-8"))
        for item in data["items"]:
            k.load_item(item)
    else:
        for ys in kalc.misc.util.split_yamldumps(data):
            k.load(ys)
    
    k._build_state()
    global kalc_state_objects
    kalc_state_objects.clear()
    kalc_state_objects.extend(k.state_objects)
    global cluster
    cluster = next(filter(lambda x: isinstance(x, GlobalVar), k.state_objects))

def run():
    # TODO HERE: copy state_objects!
    # as we will be running multiple times, we need to store original state
    # or we actually don't! we can continue computing on top of previous...?
    # for now it is ok..
    kube = KubernetesModel(kalc_state_objects)
    policy_added = False
    hypotheses = []
    for ob in kalc_state_objects:
        if isinstance(ob.policy, str): continue # STUB. find and fix
        hypotheses.append(ob.policy.apply(kube))
    # TODO HERE: generate different combinations of hypotheses
    kube.run(timeout=999000, sessionName="kalc")
    # TODO. STUB
    # TODO example hanlers and patches
    for obj in kalc_state_objects:
        if isinstance(obj, Deployment):
            if "redis-slave" in str(obj.metadata_name):
                obj.affinity_required_handler()
                # obj.scale_replicas_handler(random.randint(4,10))

    if policy_added: patch()
    for a in kube.plan:
        print(a)
        r = a()
        if isinstance(r, dict) and "kubectl" in r:
            print(">>", r["kubectl"])
    # print summary

def patch():
    for obj in kalc_state_objects:
        if isinstance(obj, Deployment):
            # print("patch for ", obj.metadata_name)
            print(highlight(obj.get_patch(), DiffLexer(), TerminalFormatter()))

def apply():
    pass


def print_objects(objectList):
    print("<==== Domain Object List =====>")

    pod_loaded_list = filter(lambda x: isinstance(x, Pod), objectList)
    print("----------Pods---------------")
    for poditem in pod_loaded_list:
        print("## Pod:"+ str(poditem.metadata_name._get_value()) + \
        ", Status: " + str(poditem.status._get_value()) + \
        ", Priority_class: " + str(poditem.priorityClass._property_value.metadata_name) + \
        ", CpuRequest: " + str(poditem.cpuRequest._get_value()) + \
        ", MemRequest: " + str(poditem.memRequest._get_value()) + \
        ", CpuLimit: " + str(poditem.cpuLimit._get_value()) + \
        ", MemLimit: " + str(poditem.memLimit._get_value()) + \
        ", ToNode: " + str(poditem.toNode._property_value) + \
        ", AtNode: " + str(poditem.atNode._property_value) + \
        ", Metadata_labels:" + str([str(x) for x in poditem.metadata_labels._property_value]) + \
        ", hasService: " + str(poditem.hasService._get_value()) + \
        ", hasDeployment: " + str(poditem.hasDeployment._get_value()) + \
        ", hasDaemonset: " + str(poditem.hasDaemonset._get_value()))
    
    node_loaded_list = filter(lambda x: isinstance(x, Node), objectList)
    print("----------Nodes---------------")
    for nodeitem in node_loaded_list:
        print("## Node:"+ str(nodeitem.metadata_name._get_value()) + \
        ", cpuCapacity: " + str(nodeitem.cpuCapacity._get_value()) + \
        ", memCapacity: " + str(nodeitem.memCapacity._get_value()) + \
        ", CurrentFormalCpuConsumption: "  + str(nodeitem.currentFormalCpuConsumption._get_value()) + \
        ", CurrentFormalMemConsumption: " + str(nodeitem.currentFormalMemConsumption._get_value()) + \
        ", AmountOfPodsOverwhelmingMemLimits: " + str(nodeitem.AmountOfPodsOverwhelmingMemLimits._get_value()) + \
        ", PodAmount: "  + str(nodeitem.podAmount._get_value()) + \
        ", IsNull:"  + str(nodeitem.isNull._get_value()) + \
        ", Status:"  + str(nodeitem.status._get_value()) +\
        ", AmountOfActivePods: " + str(nodeitem.amountOfActivePods._get_value()) +\
        ", Searchable: " + str(nodeitem.searchable._get_value())+\
        ", IsSearched: ", str(nodeitem.isSearched._get_value())+\
        ", different_than: ", str([str(x) for x in nodeitem.different_than._get_value()]))
    services = filter(lambda x: isinstance(x, Service), objectList)
    print("----------Services---------------")
    for service in services:
        print("## Service: "+str(service.metadata_name)+\
        ", AmountOfActivePods: "+str(service.amountOfActivePods._get_value())+\
        ", Status: " + str(service.status._get_value()) +
        ", Spec_selector: "+str([str(x) for x in service.spec_selector._property_value])+\
        ", Pod_List: "+str([str(x) for x in service.podList._get_value()])+\
        ", IsSearched: ", str(service.isSearched._get_value()))


    prios = filter(lambda x: isinstance(x, PriorityClass), objectList)
    print("----------PriorityClasses---------------")
    for prio in prios:
        print("## PriorityClass: "+str(prio.metadata_name) +" " + str(prio.priority._get_value()))


    scheduler = next(filter(lambda x: isinstance(x, Scheduler), objectList))
    print("----------Shedulers---------------")
    print("## Sheduler: "+str(scheduler.status._get_value()) +\
        " PodList: "+str([str(x) for x in scheduler.podQueue._get_value()]) +\
        " QueueLength: "+str(scheduler.queueLength._get_value()))

    deployments_loaded_list = filter(lambda x: isinstance(x, Deployment), objectList)
    print("----------Deployments------------")
    for deployment in deployments_loaded_list:
        print("## Deployment: "+str(deployment.metadata_name._get_value()) +\
        " Spec_replicas: "+ str(deployment.spec_replicas._get_value()) +\
        " Namespace: " + str(deployment.metadata_namespace._get_value())+\
        " AmountOfActivePods: " + str(deployment.amountOfActivePods._get_value())+\
        " Status: " + str(deployment.status._get_value())+\
        " PodList: " + str([str(x) for x in deployment.podList._get_value()])+\
        " PriorityClassName: " + str(deployment.spec_template_spec_priorityClassName._property_value) + \
        " Searchable:" + str(deployment.searchable))
        # " Metadata_labels: " + str([str(x) for x in deployment.template_metadata_labels._property_value]))
    
    daemonsets_loaded_list = filter(lambda x: isinstance(x, DaemonSet), objectList)
    print("----------DaemonSets------------")
    for daemonset in daemonsets_loaded_list:
        print("## DaemonSet: "+str(daemonset.metadata_name._get_value()) +\
        " AmountOfActivePods: " + str(daemonset.amountOfActivePods._get_value())+\
        " Status: " + str(daemonset.status._get_value())+\
        " PodList: " + str([str(x) for x in daemonset.podList._get_value()])+\
        " PriorityClassName: " + str(daemonset.spec_template_spec_priorityClassName._property_value) + \
        " Searchable:" + str(daemonset.searchable))
        # " Metadata_labels: " + str([str(x) for x in deployment.template_metadata_labels._property_value]))

    replicasets_loaded_list = filter(lambda x: isinstance(x, ReplicaSet), objectList)
    print("----------ReplicaSets------------")
    for replicaset in replicasets_loaded_list:
        print("## Replicaset: "+str(replicaset.metadata_name._get_value()) +\
        " hash: " + str(replicaset.hash)+\
        " spec_replicas: " + str(replicaset.spec_replicas._get_value())+\
        " metadata_ownerReferences__kind: " + str(replicaset.metadata_ownerReferences__name._property_value)+\
        " metadata_ownerReferences__name: " + str(replicaset.metadata_ownerReferences__name._property_value))

    globalvar_loaded_list = filter(lambda x: isinstance(x, GlobalVar), objectList)
    print("----------GlobalVar------------")
    list_of_objects_output =['']
    for globalvar_item in globalvar_loaded_list:
        list_of_objects_output.extend(['is_service_disrupted',str(globalvar_item.is_service_disrupted._get_value())])
        list_of_objects_output.extend(['is_deployment_disrupted',str(globalvar_item.is_deployment_disrupted._get_value())])
        list_of_objects_output.extend(['is_daemonset_disrupted',str(globalvar_item.is_daemonset_disrupted._get_value())])
        list_of_objects_output.extend(['is_node_disrupted',str(globalvar_item.is_node_disrupted._get_value())])
    print(list_of_objects_output)